__all__ = ["intersection"]
__doc__ = """
This module finds intersection of two curvature in 2D

Code originated from Sukhbinder's intersect.py.
The code is modified to support numba
"""

import numpy as np
from numba import njit

@njit(cache=True)
def _rect_inter_inner(x1, x2):
    n1 = x1.shape[0]-1
    n2 = x2.shape[0]-1
    x1min = np.empty((n1,n2))
    x1max = np.empty((n1,n2))
    x2min = np.empty((n1,n2))
    x2max = np.empty((n1,n2))
    for i in range(n1):
        x1min[i,:] = np.minimum(x1[i], x1[i+1])
        x1max[i,:] = np.maximum(x1[i], x1[i+1])
    for i in range(n2):
        x2min[:,i] = np.minimum(x2[i], x2[i+1])
        x2max[:,i] = np.maximum(x2[i], x2[i+1])
    
    return x1min, x2max, x1max, x2min

@njit(cache=True)
def intersection(position_collection1, position_collection2):
    """
    Computes 2D-coordinate of where two position_collection intersect.
    The position_collection is expected to be from PyElastica rod.
    """
    x1 = position_collection1[0,:]
    y1 = position_collection1[1,:]
    x2 = position_collection2[0,:]
    y2 = position_collection2[1,:]

    S1, S2, S3, S4 = _rect_inter_inner(x1, x2)
    S5, S6, S7, S8 = _rect_inter_inner(y1, y2)

    C1 = np.less_equal(S1, S2)
    C2 = np.greater_equal(S3, S4)
    C3 = np.less_equal(S5, S6)
    C4 = np.greater_equal(S7, S8)

    ii, jj = np.nonzero(C1 & C2 & C3 & C4)
    n = len(ii)

    AA = np.zeros((4, 4, n))
    BB = np.zeros((4, n))
    dxy1 = position_collection1[:,1:] - position_collection1[:,:-1]
    dxy2 = position_collection2[:,1:] - position_collection2[:,:-1]

    AA[0:2, 2, :] = -1
    AA[2:4, 3, :] = -1
    AA[0::2, 0, :] = dxy1[:,ii]
    AA[1::2, 1, :] = dxy2[:,jj]

    BB[0, :] = -x1[ii].ravel()
    BB[1, :] = -x2[jj].ravel()
    BB[2, :] = -y1[ii].ravel()
    BB[3, :] = -y2[jj].ravel()

    T = np.zeros((4, n))
    for i in range(n):
        #try:
        T[:, i] = np.linalg.solve(AA[:, :, i], BB[:, i])
        #except ValueError: # Cutoff
        #    T[:, i] = np.Inf

    in_range = (T[0, :] >= 0) & (T[1, :] >= 0) & (
        T[0, :] <= 1) & (T[1, :] <= 1)

    xy0 = T[2:, in_range].T
    return xy0[:, 0], xy0[:, 1]



from typing import Optional

from gymnasium import spaces, Env

from collections import defaultdict
import time

import numpy as np
from scipy.interpolate import interp1d

import elastica as el
from elastica.timestepper import extend_stepper_interface
from elastica._calculus import _isnan_check

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.envs.octopus.build import build_octopus
from gym_softrobot.utils.custom_elastica.callback_func import (
    RodCallBack,
    RigidCylinderCallBack,
)
from gym_softrobot.utils.intersection import intersection
from gym_softrobot.utils.render.post_processing import plot_video
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)


class BaseSimulator(
    el.BaseSystemCollection,
    el.Constraints,
    el.Connections,
    el.Forcing,
    el.CallBacks,
    el.Damping,
):
    pass


class FlatEnv(Env):
    """
    Description:
    Source:
    Observation:
    Actions:
    Reward:
    Starting State:
    Episode Termination:
    Solved Requirements:
    """

    metadata = {"render.modes": ["rgb_array", "human"]}

    def __init__(
        self,
        final_time=5.0,
        time_step=7.0e-5,
        recording_fps=5,
        n_elems=10,
        n_arm=8,
        n_action=3,
        config_generate_video=False,
        config_save_head_data=False,
        policy_mode="centralized",
    ):
        # Integrator type

        self.final_time = final_time
        self.time_step = time_step
        self.total_steps = int(self.final_time / self.time_step)
        self.recording_fps = recording_fps
        self.step_skip = int(1.0 / (recording_fps * time_step))

        self.n_arm = n_arm
        self.n_elems = n_elems
        self.n_seg = n_elems - 1
        self.policy_mode = policy_mode

        # Spaces
        self.n_action = n_action  # number of interpolation point (3 curvatures)
        shared_space = 13
        if policy_mode == "centralized":
            action_size = (n_arm * self.n_action,)
            action_low = np.ones(self.n_action) * (-22)
            action_high = np.ones(self.n_action) * (22)
            action_low = np.repeat(action_low, n_arm)
            action_high = np.repeat(action_high, n_arm)
            self.action_space = spaces.Box(
                action_low, action_high, shape=action_size, dtype=np.float32
            )
            self._observation_size = (
                n_arm,
                (self.n_seg + (self.n_elems + 1) * 4 + self.n_action),
            )
            self.observation_space = spaces.Dict(
                {
                    "individual": spaces.Box(
                        -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
                    ),
                    "shared": spaces.Box(
                        -np.inf, np.inf, shape=(shared_space,), dtype=np.float32
                    ),
                }
            )
        elif policy_mode == "decentralized":
            action_size = (self.n_action,)
            action_low = np.ones(action_size) * (-22)
            action_high = np.ones(action_size) * (22)
            self.action_space = spaces.Box(
                action_low, action_high, shape=action_size, dtype=np.float32
            )
            self._observation_size = (
                (self.n_seg + (self.n_elems + 1) * 4 + self.n_action + n_arm),
            )
            self.observation_space = spaces.Dict(
                {
                    "individual": spaces.Box(
                        -np.inf, np.inf, shape=self._observation_size, dtype=np.float32
                    ),
                    "shared": spaces.Box(
                        -np.inf, np.inf, shape=(shared_space,), dtype=np.float32
                    ),
                }
            )
        else:
            raise NotImplementedError
        self.metadata = {}
        self.reward_range = 100.0
        if policy_mode == "centralized":
            self._prev_action = np.zeros(
                list(self.action_space.shape), dtype=self.action_space.dtype
            )
        elif policy_mode == "decentralized":
            self._prev_action = np.zeros(
                [n_arm] + list(self.action_space.shape), dtype=self.action_space.dtype
            )
        else:
            raise NotImplementedError

        # Configurations
        self.config_generate_video = config_generate_video
        self.config_save_head_data = config_save_head_data

        # Rendering-related
        self.viewer = None
        self.renderer = None

    def summary(
        self,
    ):
        print(
            f"""
        {self.final_time=}
        {self.time_step=}
        {self.total_steps=}
        {self.step_skip=}
        simulation time per action: {1.0/self.step_skip=}
        max number of action per episode: {self.total_steps / self.step_skip}

        {self.n_elems=}
        {self.action_space=}
        {self.observation_space=}
        {self.reward_range=}
        """
        )

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        self.simulator = BaseSimulator()

        self.shearable_rods, self.rigid_rod = build_octopus(
            self.simulator,
            self.n_arm,
            self.n_elems,
            self.time_step,
        )

        # CallBack
        if self.config_generate_video:
            self.rod_parameters_dict_list = []
            for arm in self.shearable_rods:
                rod_parameters_dict = defaultdict(list)
                self.simulator.collect_diagnostics(arm).using(
                    RodCallBack,
                    step_skip=self.step_skip,
                    callback_params=rod_parameters_dict,
                )
                self.rod_parameters_dict_list.append(rod_parameters_dict)
        if self.config_save_head_data:
            self.head_dict = defaultdict(list)
            self.simulator.collect_diagnostics(self.rigid_rod).using(
                RigidCylinderCallBack,
                step_skip=self.step_skip,
                callback_params=self.head_dict,
            )

        """ Finalize the simulator and create time stepper """
        self.StatefulStepper = el.PositionVerlet()
        self.simulator.finalize()
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )

        # """ Return
        #     (1) total time steps for the simulation step iterations
        #     (2) systems for controller design
        # """
        # systems = [self.shearable_rod]
        self.rest_sigma = np.zeros_like(self.shearable_rods[0].sigma)
        self.time = np.float64(0.0)
        self.counter = 0
        # self.bias=self.shearable_rod.compute_position_center_of_mass()[0].copy()

        # Set Target
        self._target = (2 - 0.5) * self.np_random.random(2) + 0.5
        # self._target /= np.linalg.norm(self._target) # I don't see why this is here

        # Initial State
        state = self.get_state()

        if return_info:
            return state, {}
        else:
            return state

    def get_state(self):
        states = {}
        cx, cy, _ = self.rigid_rod.position_collection[:, 0]
        # Build state
        kappa_state = np.vstack([rod.kappa[0] for rod in self.shearable_rods])
        pos_state1 = np.vstack(
            [rod.position_collection[0] - cx for rod in self.shearable_rods]
        )  # x
        pos_state2 = np.vstack(
            [rod.position_collection[1] - cy for rod in self.shearable_rods]
        )  # y
        vel_state1 = np.vstack(
            [rod.velocity_collection[0] for rod in self.shearable_rods]
        )  # x
        vel_state2 = np.vstack(
            [rod.velocity_collection[1] for rod in self.shearable_rods]
        )  # y
        if self.policy_mode == "decentralized":
            previous_action = self._prev_action[:, :]
            individual_state = np.hstack(
                [
                    kappa_state,
                    pos_state1,
                    pos_state2,
                    vel_state1,
                    vel_state2,
                    previous_action,
                    np.eye(self.n_arm),
                ]
            ).astype(np.float32)
        elif self.policy_mode == "centralized":
            previous_action = self._prev_action.reshape([self.n_arm, self.n_action])
            individual_state = np.hstack(
                [
                    kappa_state,
                    pos_state1,
                    pos_state2,
                    vel_state1,
                    vel_state2,
                    previous_action,
                ]
            ).astype(np.float32)
        else:
            raise NotImplementedError
        shared_state = np.concatenate(
            [
                self._target - self.rigid_rod.position_collection[:2, 0],  # 2
                # self.rigid_rod.position_collection[:,0], # 3
                self.rigid_rod.velocity_collection[:2, 0],  # 2
                self.rigid_rod.director_collection[:, :, 0].ravel(),  # 9
            ],
            dtype=np.float32,
        )
        states["individual"] = individual_state
        states["shared"] = shared_state
        return states

    def set_action(self, action):
        reshaped_kappa = action.reshape((self.n_arm, self.n_action))
        if self.policy_mode == "decentralized":
            self._prev_action[:] = reshaped_kappa
        elif self.policy_mode == "centralized":
            self._prev_action[:] = reshaped_kappa.reshape([self.n_arm * self.n_action])
        else:
            raise NotImplementedError
        reshaped_kappa = np.concatenate(
            [
                np.zeros((self.n_arm, 1)),
                reshaped_kappa,
                np.zeros((self.n_arm, 1)),
            ],
            axis=-1,
        )
        reshaped_kappa = interp1d(
            np.linspace(0, 1, self.n_action + 2),  # added zero on the boundary
            reshaped_kappa,
            kind="cubic",
            axis=-1,
        )(np.linspace(0, 1, self.n_seg))
        for arm_i in range(self.n_arm):
            self.shearable_rods[arm_i].rest_kappa[0, :] = reshaped_kappa[arm_i]
            # self.shearable_rods[arm_i].rest_sigma[1, :] = self.rest_sigma[1,:] #rest_sigma.copy()
            # self.shearable_rods[arm_i].rest_sigma[2, :] = self.rest_sigma[2,:] #rest_sigma.copy()

    def step(self, action):
        rest_kappa = action  # alias

        """ Set intrinsic strains (set actions) """
        self.set_action(rest_kappa)

        """ Post-simulation """
        xposbefore = self.rigid_rod.position_collection[0:2, 0].copy()

        """ Run the simulation for one step """
        stime = time.perf_counter()
        for _ in range(self.step_skip):
            self.time = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time,
                self.time_step,
            )
        etime = time.perf_counter()

        """ Done is a boolean to reset the environment before episode is completed """
        terminated = False
        truncated = False
        survive_reward = 0.0
        forward_reward = 0.0
        control_panelty = 0.0  # 0.005 * np.square(rest_kappa.ravel()).mean()
        bending_energy = 0.0  # np.mean([rod.compute_bending_energy().mean() for rod in self.shearable_rods]) * 0.001
        shear_energy = (
            0.0  # sum([rod.compute_shear_energy() for rod in self.shearable_rods])
        )
        # Position of the rod cannot be NaN, it is not valid, stop the simulation
        invalid_values_condition = _isnan_check(
            np.concatenate(
                [rod.position_collection for rod in self.shearable_rods]
                + [rod.velocity_collection for rod in self.shearable_rods]
            )
        )
        arm_crossing = sum(
            [
                len(
                    intersection(
                        self.shearable_rods[i - 1].position_collection[:2, :],
                        self.shearable_rods[i].position_collection[:2, :],
                    )[0]
                )
                for i in range(self.n_arm - 1)
            ]
        )

        xposafter = self.rigid_rod.position_collection[0:2, 0]
        to_target = self._target - xposafter
        dist_to_target = np.linalg.norm(to_target)
        if invalid_values_condition == True:
            print(f" Nan detected in, exiting simulation now. {self.time=}")
            terminated = True
            truncated = True
            survive_reward = -50.0
        else:
            survive_reward = -0.02 * arm_crossing
            forward_reward = (
                dist_to_target - np.linalg.norm(self._target - xposbefore)
            ) / (self.step_skip * self.time_step)
            # forward_reward = np.dot(self.rigid_rod.velocity_collection[:2,0], to_target / dist_to_target)
            """ touched """
            if dist_to_target < 0.1:
                survive_reward = 100.0
                terminated = True

        # print(f'{self.counter=}, {etime-stime}sec, {self.time=}')
        timelimit = False
        if self.time > self.final_time:
            timelimit = True
            terminated = True

        reward = forward_reward - control_panelty + survive_reward - bending_energy
        # reward *= 10 # Reward scaling
        # print(f'  {reward=:.3f}: {forward_reward=:.8f}')
        # print(f'                 {control_panelty=:.3f}, {survive_reward=:.3f}')
        # print(f'                 {bending_energy=:.3f}, {shear_energy=:.3f}')
        # print(f'                 {dist_to_target-0.1=:.3f}')
        if terminated:
            reward -= dist_to_target - 0.1

        """ Return state:
            (1) current simulation time
            (2) current systems
            (3) a flag denotes whether the simulation runs correlectly
        """
        # systems = [self.shearable_rod]
        states = self.get_state()

        # Info
        info = {
            "time": self.time,
            "rods": self.shearable_rods,
            "body": self.rigid_rod,
            "TimeLimit.truncated": timelimit,
        }

        self.counter += 1

        return states, reward, terminated, truncated, info

    def save_data(self, filename_video, fps):
        if self.config_save_head_data:
            print("Saving data to pickle files ...", end="\r")
            np.savez(algo + "_head", **self.head_dict)
            print("Saving data to pickle files ... Done!")

        if self.config_generate_video:
            filename_video = f"save/{filename_video}"
            plot_video(
                self.rod_parameters_dict_list, filename_video, margin=0.2, fps=fps
            )

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(Session, BaseRenderer), (
                "Rendering module is not properly subclassed"
            )
            assert issubclass(Session, BaseElasticaRendererSession), (
                "Rendering module is not properly subclassed"
            )
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rods(self.shearable_rods)
            self.renderer.add_rigid_body(self.rigid_rod)
            self.renderer.add_point(self._target.tolist() + [0], 0.05)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None



from typing import Iterable
from abc import ABC, abstractmethod

from gym_softrobot.config import RendererType

from elastica.rod import RodBase


class BaseRenderer(ABC):
    """
    Renderer should contains the methods below.
    """

    @property
    @abstractmethod
    def type(self) -> RendererType:
        pass

    @abstractmethod
    def render(self):
        pass

    @abstractmethod
    def close(self):
        pass

class BaseElasticaRendererSession(ABC):
    """
    Elastica-renderer should contains the methods below
    """

    @abstractmethod
    def add_rod(self, rod: RodBase):
        pass

    def add_rods(self, rods: Iterable[RodBase]):
        # Batch collector
        for rod in rods:
            self.add_rod(rod)

    @abstractmethod
    def add_rigid_body(self):
        pass

    @abstractmethod
    def add_point(self):
        pass
    



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath('../'))

from gym_softrobot.version import VERSION

# -- Project information -----------------------------------------------------

project = 'gym-softrobot'
copyright = '2022, Seung Hyun Kim, Chia-Hsien Shih'
author = 'Seung Hyun Kim, Chia-Hsien Shih'

# The full version, including alpha/beta/rc tags
release = VERSION


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx_autodoc_typehints',
    #'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.mathjax',
    'numpydoc',
    'myst_parser',
]

myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

autodoc_default_flags = ['members',  'private-members', 'special-members',  'show-inheritance']
autosectionlabel_prefix_document = True

source_parsers = {
}
source_suffix = ['.rst', '.md']

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme'
html_title = "Gym-SoftRobot"
#html_logo = "_static/assets/logo_v1.png"
html_theme_options = {
    "path_to_docs": "docs",
    "repository_url": "https://github.com/skim0119/gym-softrobot",
    "repository_branch": "main",
    "use_issues_button": True,
    "use_repository_button": True,
    "use_edit_page_button": True,
    #"logo_only": True
}
html_css_files = ['custom.css']

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
#html_css_files = []


# -- Options for autodoc  ---------------------------------------------------
autodoc_member_order = 'bysource'

# -- Options for numpydoc ---------------------------------------------------
numpydoc_show_class_members = False



from gymnasium.envs.registration import register

from gym_softrobot.config import RendererType

""" Octopus Environment """
register(
    id="OctoFlat-v0",
    entry_point="gym_softrobot.envs.octopus:FlatEnv",
)

register(
    id="OctoFlatLite-v0",
    entry_point="gym_softrobot.envs.octopus:FlatEnv",
    kwargs=dict(n_arm=1, n_action=8),
)

register(
    id="OctoCrawl-v0",
    entry_point="gym_softrobot.envs.octopus:CrawlEnv",
)

register(
    id="OctoReach-v0",
    entry_point="gym_softrobot.envs.octopus:ReachEnv",
)

register(
    id="OctoArmSingle-v0",
    entry_point="gym_softrobot.envs.octopus:ArmSingleEnv",
)

register(
    id="OctoArmTwo-v0",
    entry_point="gym_softrobot.envs.octopus:ArmTwoEnv",
)

register(
    id="OctoArmPush-v0",
    entry_point="gym_softrobot.envs.octopus:ArmPushEnv",
)

register(
    id="OctoArmPush-v1",
    entry_point="gym_softrobot.envs.octopus:ArmPushEnv",
    kwargs=dict(mode="continuous"),
)

register(
    id="OctoArmPullWeight-v0",
    entry_point="gym_softrobot.envs.octopus:ArmPullWeightEnv",
    kwargs=dict(mode="continuous"),
)

""" Snake Environment """
register(
    id="ContinuumSnake-v0",
    entry_point="gym_softrobot.envs.snake:ContinuumSnakeEnv",
)

""" Simple Control Environment """

# """ Soft Arm Environment """
register(
    id="SoftArmTracking-v0",
    entry_point="gym_softrobot.envs.soft_arm:SoftArmTrackingEnv",
)
# register(
#    id='SoftArmTracking-v1',
#    entry_point='gym_softrobot.envs.soft_arm:SoftArmTrackingEnv',
#    kwargs=dict(game_mode=2)
# )

""" Soft Pendulum Environment """
register(
    id="SoftPendulum-v0", entry_point="gym_softrobot.envs.soft_pendulum:SoftPendulumEnv"
)

""" Global Configuration Parameters """
RENDERER_CONFIG = RendererType.POVRAY



"""
Created on Sep. 23, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.mplot3d import Axes3D

from gym_softrobot.utils.actuation.frames.frame import FrameBase
from gym_softrobot.utils.actuation.frames.frame_tools import (
    rod_color, 
    default_label_fontsize,
    process_position,
    process_director
)

class RodFrame(FrameBase):
    def __init__(self, figure_name, folder_name, fig_dict, gs_dict, **kwargs):
        FrameBase.__init__(
            self,
            figure_name=figure_name,
            folder_name=folder_name,
            fig_dict=fig_dict
        )
        self.gs_dict = gs_dict
        self.ax3d_flag = kwargs.get("ax3d_flag", False)
        self.plot_rod = (
            self.plot_rod3d if self.ax3d_flag else self.plot_rod2d
        )

        self.fontsize = kwargs.get("fontsize", default_label_fontsize)
        self.rod_color = kwargs.get("rod_color", rod_color)
        self.offset = kwargs.get("offset", np.zeros(3))
        self.rotation = kwargs.get("rotation", np.identity(3))
        self.reference_total_length = 1
        self.reference_configuration_flag = False
        self.set_n_elems(kwargs.get("n_elems", 100))

    def set_n_elems(self, n_elems):
        self.n_elems = n_elems
        self.s = np.linspace(0, 1, self.n_elems+1)
        self.s_shear = (self.s[:-1] + self.s[1:])/2
        self.s_kappa = self.s[1:-1].copy()

    def reset(self,):
        FrameBase.reset(self,)
        self.gs = gridspec.GridSpec(
            figure=self.fig,
            **self.gs_dict
        )
        self.axes_shear = []
        self.axes_kappa = []
        for i in range(3):
            self.axes_kappa.append(
                self.fig.add_subplot(self.gs[i, 5], xlim=[-0.1, 1.1])
            )
            self.axes_shear.append(
                self.fig.add_subplot(self.gs[i, 4], xlim=[-0.1, 1.1])
            )
        if self.ax3d_flag:
            self.ax_rod = self.fig.add_subplot(
                self.gs[0:3, 0:3], projection='3d'
            )
        else:
            self.ax_rod = self.fig.add_subplot(self.gs[0:3, 0:3])
        
        if self.reference_configuration_flag:
            self.plot_ref_configuration()
        
    def set_ref_configuration(self, position, shear, kappa):
        self.reference_position = position.copy()
        self.reference_shear = shear.copy()
        self.reference_kappa = kappa.copy()
        
        reference_length = np.linalg.norm(
            position[:, 1:]-position[:, :-1], axis=0
        )
        self.reference_total_length = reference_length.sum()
        self.set_n_elems(reference_length.shape[0])
        self.reference_configuration_flag = True
        return self.reference_total_length

    def plot_ref_configuration(self,):
        line_position = process_position(
            self.reference_position,
            self.offset, self.rotation
        ) / self.reference_total_length
        
        if self.ax3d_flag:
            self.ax_rod.plot(
                line_position[0], line_position[1], line_position[2],
                color="grey", linestyle="--"
            )
        else:
            self.ax_rod.plot(
                line_position[0], line_position[1],
                color="grey", linestyle="--"
            )

        for index_i in range(3):
            self.axes_shear[index_i].plot(
                self.s_shear,
                self.reference_shear[index_i],
                color="grey",
                linestyle="--"
            )
            self.axes_kappa[index_i].plot(
                self.s_kappa,
                self.reference_kappa[index_i],
                color="grey",
                linestyle="--"
            )

    def calculate_line_position(self, position, director, radius):
        line_center = process_position(
            position, self.offset, self.rotation
        ) / self.reference_total_length
        line_position = process_position(
            (position[:, :-1] + position[:, 1:])/2,
            self.offset, self.rotation
        )
        line_director = process_director(director, self.rotation)
        line_up = (
            (line_position + line_director[1, :, :] * radius) 
            / self.reference_total_length
        )
        line_down = (
            (line_position - line_director[1, :, :] * radius) 
            / self.reference_total_length
        )
        line_left = (
            (line_position + line_director[0, :, :] * radius)
            / self.reference_total_length
        )
        line_right = (
            (line_position - line_director[0, :, :] * radius)
            / self.reference_total_length
        )
        return line_center, [line_up, line_right, line_down, line_left]


    def plot_rod2d(self, position, director, radius, color=None):
        line_center, lines = self.calculate_line_position(
            position, director, radius
        )
        self.ax_rod.plot(
            lines[0][0], lines[0][1],
            color=self.rod_color if color is None else color
        )
        self.ax_rod.plot(
            lines[2][0], lines[2][1],
            color=self.rod_color if color is None else color
        )
        self.ax_rod.plot(
            lines[1][0], lines[1][1],
            alpha=0.2,
            color=self.rod_color if color is None else color
        )
        self.ax_rod.plot(
            lines[3][0], lines[3][1],
            alpha=0.2,
            color=self.rod_color if color is None else color
        )
        self.ax_rod.plot(
            [lines[0][0, -1], line_center[0, -1], lines[2][0, -1]],
            [lines[0][1, -1], line_center[1, -1], lines[2][1, -1]],
            color=self.rod_color if color is None else color,
            label='sim'
        )
        self.ax_rod.plot(
            [lines[1][0, -1], line_center[0, -1], lines[3][0, -1]],
            [lines[1][1, -1], line_center[1, -1], lines[3][1, -1]],
            alpha=0.2,
            color=self.rod_color if color is None else color
        )
        return self.ax_rod

    def plot_rod3d(self, position, director, radius, color=None):
        line_center, lines = self.calculate_line_position(
            position, director, radius
        )
        self.ax_rod.plot(
            line_center[0], line_center[1], line_center[2],
            color=self.rod_color if color is None else color,
            linestyle="--"
        )
        for line in lines:    
            self.ax_rod.plot(
                line[0], line[1], line[2],
                color=self.rod_color if color is None else color
            ) 
        self.ax_rod.plot(
            [lines[0][0, -1], line_center[0, -1], lines[2][0, -1]],
            [lines[0][1, -1], line_center[1, -1], lines[2][1, -1]],
            [lines[0][2, -1], line_center[2, -1], lines[2][2, -1]],
            color=self.rod_color if color is None else color,
            label='sim'
        )
        self.ax_rod.plot(
            [lines[1][0, -1], line_center[0, -1], lines[3][0, -1]],
            [lines[1][1, -1], line_center[1, -1], lines[3][1, -1]],
            [lines[1][2, -1], line_center[2, -1], lines[3][2, -1]],
            color=self.rod_color if color is None else color
        )
        return self.ax_rod

    def plot_strains(self, shear, kappa, color=None):
        for index_i in range(3):
            self.axes_shear[index_i].plot(
                self.s_shear,
                shear[index_i],
                color=self.rod_color if color is None else color
            )
            self.axes_kappa[index_i].plot(
                self.s_kappa,
                kappa[index_i],
                color=self.rod_color if color is None else color
            )
        return self.axes_shear, self.axes_kappa

    def set_ax_rod_lim(
        self, 
        x_lim=[-1.1, 1.1],
        y_lim=[-1.1, 1.1],
        z_lim=[-1.1, 1.1]
    ):
        self.ax_rod.set_xlim(x_lim)
        self.ax_rod.set_ylim(y_lim)
        if self.ax3d_flag:
            self.ax_rod.set_zlim(z_lim)

    def set_ax_strains_lim(self, axes_shear_lim=None, axes_kappa_lim=None,):
        if axes_shear_lim is None:
            axes_shear_lim = [
                [-0.11, 0.11],
                [-0.11, 0.11],
                [-0.1, 2.1]
            ]
        if axes_kappa_lim is None:
            axes_kappa_lim = [
                [-110, 110],
                [-110, 110],
                [-11, 11],
            ]
        for index_i in range(3):
            shear_mean = np.average(axes_shear_lim[index_i]) if index_i != 2 else 1
            shear_log = np.floor(
                np.log10(axes_shear_lim[index_i][1] - shear_mean)
            )
            kappa_mean = np.average(axes_kappa_lim[index_i])
            kappa_log = np.floor(
                np.log10(axes_kappa_lim[index_i][1] - kappa_mean)
            )
            
            self.axes_shear[index_i].set_ylim(axes_shear_lim[index_i])
            self.axes_kappa[index_i].set_ylim(axes_kappa_lim[index_i])
            self.axes_shear[index_i].ticklabel_format(
                axis='y', scilimits=(shear_log, shear_log),
                useOffset=shear_mean
            )
            self.axes_kappa[index_i].ticklabel_format(
                axis='y', scilimits=(kappa_log, kappa_log),
                useOffset=kappa_mean
            )

    def set_labels(self, time=None):
        if time is not None:
            self.ax_rod.set_title(
                "time={:.2f} [sec]".format(time), 
                fontsize=self.fontsize
            )
        
        self.ax_rod.legend()
        
        for index_i in range(3):
            self.axes_kappa[index_i].set_ylabel(
                "    d$_{}$".format(index_i+1),
                fontsize=self.fontsize,
                rotation=0
            )
            self.axes_kappa[index_i].yaxis.set_label_position("right")

        self.axes_shear[0].set_title("shear", fontsize=self.fontsize)
        ylim = self.axes_shear[2].get_ylim()
        ylim_mean = np.average(ylim)
        position = 0.9 * (ylim[1] - ylim_mean) + ylim_mean
        self.axes_shear[2].text(
            0, position, 'stretch', 
            fontsize=self.fontsize, 
            ha='left', va='top'
        )
        self.axes_shear[2].set_xlabel("$s$", fontsize=self.fontsize)
        
        self.axes_kappa[0].set_title("curvature", fontsize=self.fontsize)
        ylim = self.axes_kappa[2].get_ylim()
        ylim_mean = np.average(ylim)
        position = 0.9 * (ylim[1] - ylim_mean) + ylim_mean
        self.axes_kappa[2].text(
            1, position, 'twist',
            fontsize=self.fontsize,
            ha='right', va='top'
        )
        self.axes_kappa[2].set_xlabel("$s$", fontsize=self.fontsize)



from elastica import *
from elastica.joint import FreeJoint

import numpy as np
from numba import njit


@njit(cache=True)
def z_rotation(vector, theta):
    theta = theta / 180.0 * np.pi
    R = np.array(
        [
            [np.cos(theta), -np.sin(theta), 0.0],
            [np.sin(theta), np.cos(theta), 0],
            [0.0, 0.0, 1.0],
        ]
    )
    return np.dot(R, vector.T).T


class FixedJoint2Rigid(FreeJoint):
    """
    The fixed joint class restricts the relative movement and rotation
    between two nodes and elements by applying restoring forces and torques.
    For implementation details, refer to Zhang et al. Nature
    Communications (2019).
    """

    def __init__(self, k, nu, kt, angle, radius):
        """

        Parameters
        ----------
        k: float
            Stiffness coefficient of the joint.
        nu: float
            Damping coefficient of the joint.
        kt: float
            Rotational stiffness coefficient of the joint.
        """
        super().__init__(k, nu)
        # additional in-plane constraint through restoring torque
        # stiffness of the restoring constraint -- tuned empirically
        self.kt = kt
        self.angle = angle
        self.radius = radius

    # Apply force is same as free joint
    def apply_forces(self, rod_one, index_one, rod_two, index_two):
        self.rigid_rod_pos = rod_one.position_collection[..., index_one].copy()
        self.rigid_rod_pos[2] = 0.0
        self.rigid_rod_connection_dir, contact_force = self._apply_forces(
            index_one,
            index_two,
            self.rigid_rod_pos,
            rod_two.position_collection,
            rod_one.velocity_collection,
            rod_two.velocity_collection,
            rod_one.director_collection[1, ...].T,
            rod_one.external_forces,
            rod_two.external_forces,
            self.k,
            self.nu,
            self.kt,
            self.angle,
            self.radius,
        )

    @staticmethod
    @njit(cache=True)
    def _apply_forces(
        index_one,
        index_two,
        rigid_rod_pos,
        rod_two_position,
        rod_one_velocity,
        rod_two_velocity,
        rod_one_binormal,
        rod_one_external_forces,
        rod_two_external_forces,
        k,
        nu,
        kt,
        angle,
        radius,
    ):
        # return super().apply_forces(rod_one, index_one, rod_two, index_two)
        rigid_rod_connection_dir = -z_rotation(rod_one_binormal, angle)
        # rigid_rod_connection_dir=-z_rotation(rod_one.binormal.T,angle)

        rigid_rod_connection_pt = rigid_rod_connection_dir * radius
        rigid_rod_pos += rigid_rod_connection_pt[0]
        end_distance_vector = rod_two_position[..., index_two] - rigid_rod_pos
        # Calculate norm of end_distance_vector
        # this implementation timed: 2.48 µs ± 126 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)
        end_distance = np.sqrt(np.dot(end_distance_vector, end_distance_vector))

        # Below if check is not efficient find something else
        # We are checking if end of rod1 and start of rod2 are at the same point in space
        # If they are at the same point in space, it is a zero vector.
        if end_distance <= np.finfo(np.float64).eps * 1e4:
            normalized_end_distance_vector = np.array([0.0, 0.0, 0.0])
        else:
            normalized_end_distance_vector = end_distance_vector / end_distance

        elastic_force = k * end_distance_vector

        relative_velocity = (
            rod_two_velocity[..., index_two] - rod_one_velocity[..., index_one]
        )
        normal_relative_velocity = (
            np.dot(relative_velocity, normalized_end_distance_vector)
            * normalized_end_distance_vector
        )
        damping_force = -nu * normal_relative_velocity

        contact_force = elastic_force + damping_force

        rod_one_external_forces[..., index_one] += contact_force
        rod_two_external_forces[..., index_two] -= contact_force

        return rigid_rod_connection_dir, contact_force

    def apply_torques(self, rod_one, index_one, rod_two, index_two):
        # self._apply_hard_director_boundary(
        #        index_one,
        #        index_two,
        #        rod_one.director_collection,
        #        rod_two.director_collection,
        #    )
        self._apply_torques(
            index_one,
            index_two,
            self.rigid_rod_pos,
            self.rigid_rod_connection_dir,
            rod_two.position_collection,
            rod_one.director_collection,
            rod_two.director_collection,
            rod_two.rest_lengths,
            rod_one.external_torques,
            rod_two.external_torques,
            self.k,
            self.nu,
            self.kt,
            self.angle,
            self.radius,
        )

    @staticmethod
    @njit(cache=True)
    def _apply_hard_director_boundary(
        index_one,
        index_two,
        rod_one_director_collection,
        rod_two_director_collection,
    ):
        pass

    @staticmethod
    @njit(cache=True)
    def _apply_torques(
        index_one,
        index_two,
        rigid_rod_pos,
        rigid_rod_connection_dir,
        rod_two_position,
        rod_one_director_collection,
        rod_two_director_collection,
        rod_two_rest_lengths,
        rod_one_external_torques,
        rod_two_external_torques,
        k,
        nu,
        kt,
        angle,
        radius,
    ):
        # current direction of the first element of link two
        # also NOTE: - rod two is fixed at first element
        link_direction = (
            rod_two_position[..., index_two + 1] - rod_two_position[..., index_two]
        )

        # To constrain the orientation of link two, the second node of link two should align with
        # the direction of link one. Thus, we compute the desired position of the second node of link two
        # as check1, and the current position of the second node of link two as check2. Check1 and check2
        # should overlap.

        # tgt_destination = (
        #     rod_one.position_collection[..., index_one]
        #     + rod_two.rest_lengths[index_two] * rod_one.tangents[..., index_one]
        # )  # dl of rod 2 can be different than rod 1 so use rest length of rod 2

        tgt_destination = (
            rigid_rod_pos + rod_two_rest_lengths[index_two] * rigid_rod_connection_dir
        )

        curr_destination = rod_two_position[
            ..., index_two + 1
        ]  # second element of rod2

        # Compute the restoring torque
        forcedirection = -kt * (
            curr_destination - tgt_destination
        )  # force direction is between rod2 2nd element and rod1
        torque = np.cross(link_direction, forcedirection)[0]

        # The opposite torque will be applied on link one
        for i in range(3):
            for j in range(3):
                rod_one_external_torques[i, index_one] -= (
                    rod_one_director_collection[i, j, index_one] * torque[j]
                )
                rod_two_external_torques[i, index_two] += (
                    rod_two_director_collection[i, j, index_two] * torque[j]
                )
        # rod_one_external_torques[..., index_one] -= (
        #    rod_one_director_collection[..., index_one] @ torque
        # )
        # rod_two_external_torques[..., index_two] += (
        #    rod_two_director_collection[..., index_two] @ torque
        # )



from typing import Optional, Iterable

import numpy as np

import matplotlib
matplotlib.use("Agg")  # Must be before importing matplotlib.pyplot or pylab!
from matplotlib import pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.colors import to_rgb
from mpl_toolkits.mplot3d import proj3d, Axes3D

from abc import ABC, abstractmethod

from gym_softrobot.config import RendererType
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)

import pkg_resources

def render_figure(fig:plt.figure):
    w, h = fig.get_size_inches()
    dpi_res = fig.get_dpi()
    w, h = int(np.ceil(w * dpi_res)), int(np.ceil(h*dpi_res))

    canvas = FigureCanvasAgg(fig)
    canvas.draw()
    data = np.asarray(canvas.buffer_rgba())[:,:,:3]
    return data

def convert_marker_size(radius, ax):
    """
    Convert marker size from radius to s (in scatter plot).

    Parameters
    ----------
    radius : np.array or float
        Array (or a number) of radius
    ax : matplotlib.Axes
    """
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    max_axis_length = max(abs(xlim[1]-xlim[0]), abs(ylim[1]-ylim[0]))
    scaling_factor = 3.0e3 * (2*0.1) / max_axis_length
    return np.sqrt(np.pi * (scaling_factor * radius))
    #ppi = 72 # standard point size in matplotlib is 72 points per inch (ppi), no matter the dpi
    #point_whole_ax = 5 * 0.8 * ppi
    #point_radius= 2 * radius / 1.0 * point_whole_ax
    #return point_radius**2

def set_axes_equal(ax):
    '''Make axes of 3D plot have equal scale so that spheres appear as spheres,
    cubes as cubes, etc..  This is one possible solution to Matplotlib's
    ax.set_aspect('equal') and ax.axis('equal') not working for 3D.

    Input
      ax: a matplotlib axis, e.g., as output from plt.gca().
    '''

    x_limits = ax.get_xlim3d()
    y_limits = ax.get_ylim3d()
    z_limits = ax.get_zlim3d()

    x_range = abs(x_limits[1] - x_limits[0])
    x_middle = np.mean(x_limits)
    y_range = abs(y_limits[1] - y_limits[0])
    y_middle = np.mean(y_limits)
    z_range = abs(z_limits[1] - z_limits[0])
    z_middle = np.mean(z_limits)

    # The plot bounding box is a sphere in the sense of the infinity
    # norm, hence I call half the max range the plot radius.
    plot_radius = 0.5*max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])

class Geom(ABC):
    @abstractmethod
    def __call__(self):
        pass


class ElasticaRod(Geom):
    # RGB color must be 2d array 
    rgb_color = np.array([[0.35, 0.29, 1.0]])

    def __init__(self, rod, ax):
        self.rod = rod
        self.ax = ax

        # Initialize scatter plot
        pos, rad = self.get_position_radius()
        self.scatter = ax.scatter(pos[0,:], pos[1,:], pos[2,:], s=convert_marker_size(rad, ax), c=ElasticaRod.rgb_color)

    def get_position_radius(self):
        pos = self.rod.position_collection.copy()
        rad = self.rod.radius.copy()
        rad /= 2
        if not pos.shape[-1] == rad.shape[0]:
            # radius defined at element, while position defined at node.
            # typical elastica has n_node = n_elem + 1 (unless the rod is circular)
            pos = 0.5 * (pos[..., 1:] + pos[..., :-1])
        return pos, rad

    def __call__(self):
        # Update scatter plot positions
        pos, rad = self.get_position_radius()
        self.scatter._offsets3d = tuple(pos)

        # Updater radius
        self.scatter.set_sizes(convert_marker_size(rad, self.ax))
        
        return self.scatter

class ElasticaRodDirector(Geom):
    # TODO
    def __init__(self, rod, ax):
        self.rod = rod
        self.ax = ax

    def __call__(self):
        return None


class ElasticaCylinder(Geom):
    rgb_color = np.array([[0.35, 0.29, 1.0]])

    def __init__(self, body, ax):
        self.body = body
        self.ax = ax

        # Initialize scatter plot
        pos1, pos2, rad = self.get_position_radius()
        end_caps = np.vstack((pos1, pos2))
        size = convert_marker_size(rad/2, ax)
        self.scatter = ax.scatter(end_caps[:,0], end_caps[:,1], end_caps[:,2], s=size, c=ElasticaCylinder.rgb_color)
        #self.line, = ax.plot(end_caps[:,0], end_caps[:,1], end_caps[:,2], linewidth=size**0.5, c=ElasticaCylinder.rgb_color)

    def get_position_radius(self):
        rad = self.body.radius[0]
        length = self.body.length
        tangent = self.body.director_collection[2, :, 0]
        pos1 = self.body.position_collection[:, 0]
        pos2 = pos1 + length * tangent
        return pos1, pos2, rad

    def __call__(self):
        # Update scatter plot positions
        pos1, pos2, rad = self.get_position_radius()
        end_caps = np.vstack((pos1, pos2))
        self.scatter._offsets3d = end_caps[:,0], end_caps[:,1], end_caps[:,2]

        # Update line plot positions
        #self.line.set_data(end_caps[:,0], end_caps[:,1])
        #self.line.set_3d_properties(end_caps[:,2])

        # Updater radius (rigid body)
        
        #return [self.scatter, self.line]
        return [self.scatter]


class ElasticaSphere(Geom):
    rgb_color = np.array([1.0, 0.0, 1.0])

    def __init__(self, loc, radius, ax):
        # Initialize scatter plot
        self.scatter = ax.scatter(loc[0], loc[1], loc[2], s=convert_marker_size(radius, ax), c=ElasticaSphere.rgb_color)

    def __call__(self):
        return self.scatter


class Session(BaseElasticaRendererSession, BaseRenderer):
    def __init__(self, width, height, dpi=100):
        self.object_collection = []
        self.width = width
        self.height = height
        self.dpi = dpi

        px = 1.0 / dpi
        self.fig = plt.figure(
            figsize=(width*px,height*px),
            frameon=True,
            dpi=dpi,
        )
        self.ax = plt.axes(projection="3d")
        self.ax.set_xlabel("x")
        self.ax.set_ylabel("y")
        self.ax.set_zlabel("z")

    @property
    def type(self):
        return RendererType.MATPLOTLIB

    def add_rod(self, rod):
        self.object_collection.append(ElasticaRod(rod, self.ax))
        # TODO Maybe give another configuration to plot the directors
        # self.object_collection.append(ElasticaRodDirector(rod, self.ax))

    def add_rigid_body(self, body):
        self.object_collection.append(ElasticaCylinder(body, self.ax))

    def add_point(self, loc: list, radius: float):
        # Add static sphere
        self.object_collection.append(ElasticaSphere(loc, radius, self.ax))

    def render(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        camera_param: Optional[tuple] = None, # POVray parameter
        **kwargs
    ):
        # Reset width and height
        if not width:
            width = self.width
        if not height:
            height = self.height

        # Maybe convert povray cmaera_param to matplotlib viewpoint

        objects = [obj() for obj in self.object_collection]
        self.rescale_axis()
        rendered_data = render_figure(self.fig)
        return rendered_data

    def close(self):
        plt.close(plt.gcf())
        self.object_collection.clear()

    def rescale_axis(self):
        self.ax.relim()
        self.ax.autoscale_view()
        set_axes_equal(self.ax)



import numpy as np
import numba
from numba import njit
from elastica.external_forces import NoForces
from scipy.interpolate import make_interp_spline


class MuscleTorquesWithVaryingBetaSplines(NoForces):
    """

    This class compute the muscle torques using Beta spline.
    Points of beta spline can be changed through out the simulation, and
    every time it changes a new spline generated. Control algorithm has to
    select the spline control points. Location of control points on the arm
    is fixed and they are equidistant points.

    Attributes
    ----------
    direction : str
        Depending on the user input direction, computed torques are applied in the direction of d1, d2, or d3.
    points_array : numpy.ndarray or callable object
        This variable is a reference to points_func_array variable which can be a numpy.ndarray or callable object.
    base_length : float
        Initial length of the arm.
    muscle_torque_scale : float
        Scaling factor for beta spline muscle torques. Beta spline is non-dimensional and muscle_torque_scale scales it.
    torque_profile_recorder : defaultdict(list)
        This is a dictionary to store time-history of muscle torques and beta-spline.
    step_skip : int
        Determines the data collection step.
    counter : int
        Used to determine the current call step of this object.
    number_of_control_points : int
        Number of control points used in beta spline. Note that these are the control points in the middle and there
        are two more control points at the start and end of the rod, which are 0.
    points_cached : numpy.ndarray
        2D (2, number_of_control_points+2) array containing data with 'float' type.
        This array stores the location of control points in first row and in the second row it stores the values of
        control points selected at previous step. If control points are changed, points_cached updated.
    max_rate_of_change_of_control_points : float
        This limits the maximum change that can happen for control points in between two calls of this object.
    my_spline : object
        Stores the beta spline object generated by control points.
    """

    def __init__(
        self,
        base_length,
        number_of_control_points,
        points_func_array,
        muscle_torque_scale,
        direction,
        step_skip,
        max_rate_of_change_of_activation=0.01,
        **kwargs,
    ):
        """

        Parameters
        ----------
        base_length : float
            Initial length of the arm.
        number_of_control_points : int
            Number of control points used in beta spline. Note that these are the control points in the middle and there
            are two more control points at the start and end of the rod, which are 0.
        points_func_array  : numpy.ndarray
            2D (2, number_of_control_points+2) array containing data with 'float' type.
            This array stores the location of control points in first row and in the second row it stores the values of
            control points selected at previous step. If control points are changed, points_cached updated.
        muscle_torque_scale : float
            Scaling factor for beta spline muscle torques. Beta spline is non-dimensional and muscle_torque_scale
            scales it.
        direction  : str
            Depending on the user input direction, computed torques are applied in the "normal", "binormal", "tangent".
        step_skip  : int
            Determines the data collection step.
        max_rate_of_change_of_control_points : float
            This limits the maximum change that can happen for control points in between two calls of this object.
        **kwargs
            Arbitrary keyword arguments.
        """
        super(MuscleTorquesWithVaryingBetaSplines, self).__init__()

        if direction == str("normal"):
            self.direction = int(0)
        elif direction == str("binormal"):
            self.direction = int(1)
        elif direction == str("tangent"):
            self.direction = int(2)
        else:
            raise NameError(
                "Please type normal, binormal or tangent as muscle torque direction. Input should be string."
            )

        self.points_array = (
            points_func_array
            if hasattr(points_func_array, "__call__")
            else lambda time_v: points_func_array
        )

        self.base_length = base_length
        self.muscle_torque_scale = muscle_torque_scale

        self.torque_profile_recorder = kwargs.get("torque_profile_recorder", None)
        self.step_skip = step_skip
        self.counter = 0  # for recording data from the muscles
        self.number_of_control_points = number_of_control_points
        self.points_cached = np.zeros(
            (2, self.number_of_control_points + 2)
        )  # This caches the control points. Note that first and last control points are zero.
        self.points_cached[0, :] = np.linspace(
            0, self.base_length, self.number_of_control_points + 2
        )  # position of control points along the rod.
        self.points_cached[1, 1:-1] = np.zeros(
            self.number_of_control_points
        )  # initalize at a value that RL can not match

        # Max rate of change of activation determines, maximum change in activation
        # signal in one time-step.
        self.max_rate_of_change_of_activation = max_rate_of_change_of_activation

        # Purpose of this flag is to just generate spline even the control points are zero
        # so that code wont crash.
        self.initial_call_flag = 0

    def apply_torques(self, system, time: np.float = 0.0):

        # Check if RL algorithm changed the points we fit the spline at this time step
        # if points_array changed create a new spline. Using this approach we don't create a
        # spline every time step.
        # Make sure that first and last point y values are zero. Because we cannot generate a
        # torque at first and last nodes.
        # print('torque',self.max_rate_of_change_of_activation)

        if (
            not np.array_equal(self.points_cached[1, 1:-1], self.points_array(time))
            or self.initial_call_flag == 0
        ):
            self.initial_call_flag = 1

            # Apply filter to the activation signal, to prevent drastic changes in activation signal.
            self.filter_activation(
                self.points_cached[1, 1:-1],
                np.array((self.points_array(time))),
                self.max_rate_of_change_of_activation,
            )

            self.my_spline = make_interp_spline(
                self.points_cached[0], self.points_cached[1]
            )
            cumulative_lengths = np.cumsum(system.lengths)

            # Compute the muscle torque magnitude from the beta spline.
            self.torque_magnitude_cache = self.muscle_torque_scale * self.my_spline(
                cumulative_lengths
            )

        self.compute_muscle_torques(
            self.torque_magnitude_cache,
            self.direction,
            system.external_torques,
        )

        if self.counter % self.step_skip == 0:
            if self.torque_profile_recorder is not None:
                self.torque_profile_recorder["time"].append(time)

                self.torque_profile_recorder["torque_mag"].append(
                    self.torque_magnitude_cache.copy()
                )
                self.torque_profile_recorder["torque"].append(
                    system.external_torques.copy()
                )
                self.torque_profile_recorder["element_position"].append(
                    np.cumsum(system.lengths)
                )

        self.counter += 1

    @staticmethod
    @njit(cache=True)
    def compute_muscle_torques(torque_magnitude, direction, external_torques):
        """
        This Numba function updates external torques, it is used here because it updates faster than numpy version.

        Parameters
        ----------
        torque_magnitude : numpy.ndarray
            1D (n_elem,) array containing data with 'float' type.
            Computed muscle torque values.
        direction : int
            Determines which component of torque vector updated.
        external_torques : numpy.ndarray
            2D (3, n_elem) array containing data with 'float' type.

        Returns
        -------

        """

        blocksize = torque_magnitude.shape[0]
        for k in range(blocksize):
            external_torques[direction, k] += torque_magnitude[k]

    @staticmethod
    @numba.njit()
    def filter_activation(signal, input_signal, max_signal_rate_of_change):
        """
        Filters the input signal. If change in new signal (input signal) greater than
        previous signal (signal) then, increase for signal is max_signal_rate_of_change amount.

        Parameters
        ----------
        signal : numpy.ndarray
            1D (number_of_control_points,) array containing data with 'float' type.
        input_signal : numpy.ndarray
            1D (number_of_control_points,) array containing data with 'float' type.
        max_signal_rate_of_change : float
            This limits the maximum change that can happen between signal and input signal.

        Returns
        -------

        """
        signal_difference = input_signal - signal
        signal += np.sign(signal_difference) * np.minimum(
            max_signal_rate_of_change, np.abs(signal_difference)
        )



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

from gym_softrobot.utils.actuation.rod_tools import (
    calculate_dilatation,
    sigma_to_shear,
    kappa_to_curvature,
)

import numpy as np


class Algorithm(object):
    def __init__(self, rod, algo_config):
        
        self.rod = rod
        self.config = algo_config


        '''
        make a copy of the required info of the rod 
        so that it will not affect the values of the origin copy
        '''
        
        self.n_elems = rod.n_elems
        self.reference_length = np.sum(rod.rest_lengths)

        # self.ds = 1 / self.n_elems
        # self.dl = self.reference_length / self.n_elems
        self.dl = rod.rest_lengths.copy()
        self.ds = self.dl / self.reference_length
        self.s = np.insert(np.cumsum(self.ds), 0, 0)
        self.s_position = self.s.copy()
        self.s_director = (self.s[:-1] + self.s[1:])/2
        self.s_shear = (self.s[:-1] + self.s[1:])/2
        self.s_kappa = self.s[1:-1]

        self.position = rod.position_collection.copy()
        self.director = rod.director_collection.copy()

        # _, voronoi_dilatation = calculate_dilatation(rod.sigma)
        self.shear = sigma_to_shear(rod.sigma)
        # self.curvature = kappa_to_curvature(rod.kappa, voronoi_dilatation)

        # # self.shear = rod.sigma + np.array([0, 0, 1])[:, None]
        # # voronoi_dilatation = np.ones(rod.voronoi_dilatation.shape)
        # # self.curvature = rod.kappa / voronoi_dilatation

        # _, voronoi_dilatation = calculate_dilatation(rod.rest_sigma)
        self.rest_shear = sigma_to_shear(rod.rest_sigma)
        # self.rest_curvature = kappa_to_curvature(rod.kappa, voronoi_dilatation)

        # # self.rest_shear = rod.rest_sigma.copy()
        # # self.rest_shear[2, :] += 1
        # # self.rest_curvature = rod.rest_kappa / rod.voronoi_dilatation

        self.sigma = rod.sigma.copy()
        self.kappa = rod.kappa.copy()

        self.rest_sigma = rod.rest_sigma.copy()
        self.rest_kappa = rod.rest_kappa.copy()

        self.radius = rod.radius.copy()
        self.shear_matrix = rod.shear_matrix.copy()
        self.bend_matrix = rod.bend_matrix.copy()

        self.internal_force = np.zeros((3, self.n_elems))
        self.internal_couple = np.zeros((3, self.n_elems-1))

        # self.target = algo_config.get('target', None)
        # self.prev_target = None

        # self.s = np.linspace(0, 1, self.n_elems+1)
        # self.s_sigma = (self.s[:-1] + self.s[1:])/2
        # self.s_kappa = self.s[1:-1]

        # self.running_cost = 0
        # self.terminal_cost = 0

    def run(self, plot_flag=False):
        raise NotImplementedError



import pytest

""" Test Dependency Installation

The purpose is to check if core dependencies are installed properly.
Typically, failure to these tests indicate an incorrection installation 
or wrong activation of the virtual environment (i.e. conda, venv, etc.).

"""
def test_gym():
    import gym
    version = gym.version.VERSION
    assert version

def test_pyelastica():
    import elastica



import gym
import numpy as np
import random

from ray.rllib.env.multi_agent_env import MultiAgentEnv


class BasicMultiAgent(MultiAgentEnv):
    metadata = {
        "render.modes": ["rgb_array"],
    }

    def __init__(self, env, agent_ids, agent_state_index_list, shared_state_index_list):
        super().__init__()
        self.env = env
        self.agent_ids = agent_ids
        self.agent_state_index_list = agent_state_index_list
        self.shared_state_index_list = shared_state_index_list
        # self.agent_ids = ["joint0", "joint1"]
        # self.agent_state_index_list = [[0, 2], [1, 3]]
        # self.shared_state_index_list = [4, 5, 6, 7, 8, 9]
        self.num_agent = len(self.agent_ids)
        self.observation_space= gym.spaces.Box(low=-np.inf, high=np.inf, shape=(8,))
        self.action_space = gym.spaces.Box(low=-1, high=1, shape=(1,))
        self.dones = set()
        self.resetted = False

    def reset(self):
        obs = self.env.reset()
        self.resetted = True
        self.dones = set()
        return self.state_fn(obs)

    def state_fn(self, obs):
        return {
            self.agent_ids[i]: np.concatenate([obs[self.agent_state_index_list[i]], obs[self.shared_state_index_list]],
                                              axis=0) for i in range(self.num_agent)}

    def step(self, action_dict):
        obs, rew, done, _ = self.env.step(list(action_dict.values()))
        agent_obs = self.state_fn(obs)
        agent_rew = {self.agent_ids[i]: rew for i in range(self.num_agent)}
        agent_done = {self.agent_ids[i]: done for i in range(self.num_agent)}
        agent_done["__all__"] = done
        return agent_obs, agent_rew, agent_done, {}

    def render(self, mode="rgb_array"):
        # Just generate a random image here for demonstration purposes.
        # Also see `gym/envs/classic_control/cartpole.py` for
        # an example on how to use a Viewer object.
        return np.random.randint(0, 256, size=(200, 300, 3), dtype=np.uint8)




import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import to_rgb
from tqdm import tqdm


def plot_snake_velocity(
    plot_params: dict,
    period,
    filename="slithering_snake_velocity.png",
    SAVE_FIGURE=False,
):
    time_per_period = np.array(plot_params["time"]) / period
    avg_velocity = np.array(plot_params["avg_velocity"])

    [
        velocity_in_direction_of_rod,
        velocity_in_rod_roll_dir,
        _,
        _,
    ] = compute_projected_velocity(plot_params, period)

    fig = plt.figure(figsize=(10, 8), frameon=True, dpi=150)
    plt.rcParams.update({"font.size": 16})
    ax = fig.add_subplot(111)
    ax.grid(b=True, which="minor", color="k", linestyle="--")
    ax.grid(b=True, which="major", color="k", linestyle="-")
    ax.plot(
        time_per_period[:], velocity_in_direction_of_rod[:, 2], "r-", label="forward"
    )
    ax.plot(
        time_per_period[:],
        velocity_in_rod_roll_dir[:, 0],
        c=to_rgb("xkcd:bluish"),
        label="lateral",
    )
    ax.plot(time_per_period[:], avg_velocity[:, 1], "k-", label="normal")
    ax.set_ylabel("Velocity [m/s]", fontsize=16)
    ax.set_xlabel("Time [s]", fontsize=16)
    fig.legend(prop={"size": 20})
    plt.show()
    # Be a good boy and close figures
    # https://stackoverflow.com/a/37451036
    # plt.close(fig) alone does not suffice
    # See https://github.com/matplotlib/matplotlib/issues/8560/
    plt.close(plt.gcf())

    if SAVE_FIGURE:
        fig.savefig(filename)


def plot_video(
    plot_params: dict,
    video_name="video.mp4",
    fps=15,
    xlim=(0, 4),
    ylim=(-1, 1),
):  # (time step, x/y/z, node)
    import matplotlib.animation as manimation

    positions_over_time = np.array(plot_params["position"])

    print("plot video")
    FFMpegWriter = manimation.writers["ffmpeg"]
    metadata = dict(title="Movie Test", artist="Matplotlib", comment="Movie support!")
    writer = FFMpegWriter(fps=fps, metadata=metadata)
    fig = plt.figure(figsize=(10, 8), frameon=True, dpi=150)
    ax = fig.add_subplot(111)
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    ax.set_xlabel("z [m]", fontsize=16)
    ax.set_ylabel("x [m]", fontsize=16)
    rod_lines_2d = ax.plot(positions_over_time[0][2], positions_over_time[0][0])[0]
    # plt.axis("equal")
    with writer.saving(fig, video_name, dpi=150):
        for time in tqdm(range(1, len(plot_params["time"]))):
            rod_lines_2d.set_xdata(positions_over_time[time][2])
            rod_lines_2d.set_ydata(positions_over_time[time][0])
            writer.grab_frame()

    # Be a good boy and close figures
    # https://stackoverflow.com/a/37451036
    # plt.close(fig) alone does not suffice
    # See https://github.com/matplotlib/matplotlib/issues/8560/
    plt.close(plt.gcf())


def compute_projected_velocity(plot_params: dict, period):

    time_per_period = np.array(plot_params["time"]) / period
    avg_velocity = np.array(plot_params["avg_velocity"])
    center_of_mass = np.array(plot_params["center_of_mass"])

    # Compute rod velocity in rod direction. We need to compute that because,
    # after snake starts to move it chooses an arbitrary direction, which does not
    # have to be initial tangent direction of the rod. Thus we need to project the
    # snake velocity with respect to its new tangent and roll direction, after that
    # we will get the correct forward and lateral speed. After this projection
    # lateral velocity of the snake has to be oscillating between + and - values with
    # zero mean.

    # Number of steps in one period.
    period_step = int(1.0 / (time_per_period[-1] - time_per_period[-2]))
    number_of_period = int(time_per_period[-1])

    # Center of mass position averaged in one period
    center_of_mass_averaged_over_one_period = np.zeros((number_of_period - 2, 3))
    for i in range(1, number_of_period - 1):
        # position of center of mass averaged over one period
        center_of_mass_averaged_over_one_period[i - 1] = np.mean(
            center_of_mass[(i + 1) * period_step : (i + 2) * period_step]
            - center_of_mass[(i + 0) * period_step : (i + 1) * period_step],
            axis=0,
        )
    # Average the rod directions over multiple periods and get the direction of the rod.
    direction_of_rod = np.mean(center_of_mass_averaged_over_one_period, axis=0)
    direction_of_rod /= np.linalg.norm(direction_of_rod, ord=2)

    # Compute the projected rod velocity in the direction of the rod
    velocity_mag_in_direction_of_rod = np.einsum(
        "ji,i->j", avg_velocity, direction_of_rod
    )
    velocity_in_direction_of_rod = np.einsum(
        "j,i->ji", velocity_mag_in_direction_of_rod, direction_of_rod
    )

    # Get the lateral or roll velocity of the rod after subtracting its projected
    # velocity in the direction of rod
    velocity_in_rod_roll_dir = avg_velocity - velocity_in_direction_of_rod

    # Compute the average velocity over the simulation, this can be used for optimizing snake
    # for fastest forward velocity. We start after first period, because of the ramping up happens
    # in first period.
    average_velocity_over_simulation = np.mean(
        velocity_in_direction_of_rod[period_step * 2 :], axis=0
    )

    return (
        velocity_in_direction_of_rod,
        velocity_in_rod_roll_dir,
        average_velocity_over_simulation[2],
        average_velocity_over_simulation[0],
    )


def plot_curvature(
    plot_params: dict,
    rest_lengths,
    period,
    save_fig=False,
    filename="continuum_snake_curvature",
):
    s = np.cumsum(rest_lengths)
    L0 = s[-1]
    s = s / L0
    s = s[:-1].copy()
    x = np.linspace(0, 1, 100)
    curvature = np.array(plot_params["curvature"])
    time = np.array(plot_params["time"])
    peak_time = period * 0.125
    dt = time[1] - time[0]
    peak_idx = int(peak_time / (dt))
    plt.rcParams.update({"font.size": 16})
    fig = plt.figure(figsize=(10, 8), frameon=True, dpi=150)
    ax = fig.add_subplot(111)
    try:
        for i in range(peak_idx * 8, peak_idx * 8 * 2, peak_idx):
            ax.plot(s, curvature[i, 0, :] * L0, "k")
    except:
        print("Simulation time not long enough to plot curvature")
    ax.plot(
        x, 7 * np.cos(2 * np.pi * x - 0.80), "--", label="stereotypical snake curvature"
    )
    ax.set_ylabel(r"$\kappa$", fontsize=16)
    ax.set_xlabel("s", fontsize=16)
    ax.set_xlim(0, 1)
    ax.set_ylim(-10, 10)
    fig.legend(prop={"size": 16})
    plt.show()
    if save_fig:
        fig.savefig(filename)

    # Be a good boy and close figures
    # https://stackoverflow.com/a/37451036
    # plt.close(fig) alone does not suffice
    # See https://github.com/matplotlib/matplotlib/issues/8560/
    plt.close(plt.gcf())



from collections import defaultdict
import time
import copy

import numpy as np
from scipy.interpolate import interp1d

from typing import Optional
import gym
from gym import core
from gym import error, spaces, utils
from gym.utils import seeding

from elastica._calculus import _isnan_check
from elastica.timestepper import extend_stepper_interface
from elastica import *
from elastica.external_forces import GravityForces

from gym_softrobot import RENDERER_CONFIG
from gym_softrobot.config import RendererType
from gym_softrobot.envs.octopus.build import build_arm
from gym_softrobot.utils.custom_elastica.callback_func import (
    RodCallBack,
    RigidCylinderCallBack,
)
from gym_softrobot.utils.render.post_processing import plot_video
from gym_softrobot.utils.render.base_renderer import (
    BaseRenderer,
    BaseElasticaRendererSession,
)

from gym_softrobot.utils.custom_elastica.muscle_torque import (
    MuscleTorquesWithVaryingBetaSplines,
)

RENDERER_CONFIG = RendererType.MATPLOTLIB

# Set base simulator class
class BaseSimulator(BaseSystemCollection, Constraints, Connections, Forcing, CallBacks):
    pass


# TODO: generalize this as a Class for online trajectory generation.
def generate_trajectory(final_time, sim_dt, target_v_scale):
    end_time = (
        final_time * 1.1
    )  # Use of 1.1 just adds a little buffer on the end of the trajectory.
    numpoints = np.rint(1 / sim_dt * end_time).astype(int)
    target_trajectory = np.zeros((numpoints, 3))

    """ Generate a trajectory bounded in a box [-0.8, 0.8] x [-0.8, 0.8] x [0.0, 0.8] """
    t = np.array([end_time * float(i) / (numpoints - 1) for i in range(numpoints)])
    t += (
        np.random.rand(1)[0] * 3600
    )  # start the trajectory anywhere within a 1 hour period

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 0] = (
        direction
        * 0.8
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 1] = (
        direction
        * 0.4
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )
    target_trajectory[:, 1] += 0.4

    f1 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f2 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    f3 = np.random.uniform(2, 5) * 0.025 * target_v_scale
    direction = np.random.randint(0, 1) * 2 - 1
    target_trajectory[:, 2] = (
        direction
        * 0.8
        * np.sin(2 * np.pi * f1 * t)
        * np.sin(2 * np.pi * f2 * t)
        * np.sin(2 * np.pi * f3 * t)
        * 1000
    )

    return target_trajectory


class SoftArmTrackingEnv(gym.Env):
    metadata = {"render.modes": ["rgb", "human"]}

    def __init__(self, **kwargs):

        if kwargs:
            print(
                "The following default parameters are being overwritten to the displayed value:"
            )
            for key in kwargs.keys():
                print("    %s:" % (key), kwargs[key])
        print("")

        """ numerical parameters """
        self.n_elem = kwargs.get("n_elem", 40)
        self.sim_dt = kwargs.get("sim_dt", 2.0e-4)  # seconds

        """ Environment time parameters """
        self.RL_update_interval = kwargs.get(
            "RL_update_interval", 0.01
        )  # This is 100 updates per second
        self.num_steps_per_update = np.rint(
            self.RL_update_interval / self.sim_dt
        ).astype(int)

        """ Target mode and parameters """
        self.mode = kwargs.get("mode", 1)
        if self.mode == 1:
            self.target_location = kwargs.get(
                "target_location", np.array([500, 500.0, 500])
            )
            self.max_episode_final_time = kwargs.get(
                "max_episode_final_time", 5
            )  # seconds
        elif self.mode == 2:
            self.target_v_scale = kwargs.get("target_v_scale", 0.1)
            self.max_episode_final_time = kwargs.get(
                "max_episode_final_time", 20
            )  # seconds

        self.num_timesteps_per_episode = np.rint(
            self.max_episode_final_time / self.RL_update_interval
        ).astype(int)

        """ Arm parameters """
        self.base_length = kwargs.get("base_length", 1) * 1000  # m --> mm
        self.radius = kwargs.get("radius", 0.025) * 1000  # m --> mm
        self.youngs_modulus = kwargs.get(
            "youngs_modulus", 2e6
        )  # Pa -- kg/s2 m --> g/s2 mm (does not require rescaling)
        self.damping = (
            self.youngs_modulus * 1e-7 * 10
        )  # kg/m s --> g/mm s (does not require rescaling)
        self.torque_damping_ratio = (
            1 * 1e6
        )  # ratio of torque to force damping coefficient (1e6 accounts for g/mm/s unit conversion -- ratio usually defined in kg/m/s)

        if "poisson_ratio" in kwargs.keys():
            poisson_ratio = kwargs.get("poisson_ratio")
            self.shear_modulus = (
                0.5 * self.youngs_modulus / (poisson_ratio + 1.0)
            )  # Pa -- kg/s2 m --> g/s2 mm (does not require rescaling)
        elif "shear_modulus" in kwargs.keys():
            self.shear_modulus = kwargs.get(
                "shear_modulus"
            )  # Pa -- kg/s2 m --> g/s2 mm (does not require rescaling)
        else:
            poisson_ratio = 0.45  # Default is to assume a Poisson ratio of 0.45
            self.shear_modulus = (
                0.5 * self.youngs_modulus / (poisson_ratio + 1.0)
            )  # Pa -- kg/s2 m --> g/s2 mm (does not require rescaling)

        self.density = kwargs.get("density", 1000) * 1e-6  # kg/m3 --> g/mm3

        """ Add gravity? """
        self.gravity_mode = kwargs.get("gravity_mode", 0)

        """ Torque activation parameters """
        self.torque_scale = kwargs.get("torque_scale", 10)
        self.max_rate_of_change_of_activation = kwargs.get(
            "max_rate_of_change_of_activation", np.infty
        )

        self.StatefulStepper = PositionVerlet()

        """ State and action space parameters """
        self.number_of_control_points = kwargs.get("number_of_control_points", 4)
        self.number_of_observation_segments = kwargs.get(
            "number_of_observation_segments", self.number_of_control_points
        )
        # normal and/or binormal direction activation (3D)
        self.action_space = gym.spaces.Box(
            low=-1.0,
            high=1.0,
            shape=(2 * self.number_of_control_points,),
            dtype=np.float64,
        )
        self.observation_space = gym.spaces.Box(
            low=-np.inf,
            high=np.inf,
            shape=(self.number_of_observation_segments * 2 + 6,),
            dtype=np.float64,
        )

        """ Rendering and logging parameters """
        self.viewer = None
        self.renderer = None
        self.COLLECT_DATA_FOR_POSTPROCESSING = False
        self.rendering_fps = 30
        self.step_skip = np.rint(1.0 / (self.rendering_fps * self.sim_dt)).astype(int)

        #
        self.get_state = kwargs.get("custom_state_fn", self.get_state_default)

    def get_state_default(self):
        """
        Returns current state of the system to the controller.

        Returns
        -------
        numpy.ndarray
            1D (number_of_states) array containing data with 'float' type.
            Size of the states depends on the problem.
        """
        avg_kappa_1 = np.zeros(self.number_of_observation_segments)
        # self.n_elem-1 is number of curvature values due to veroni regions
        avg_length = int((self.n_elem - 1) / (self.number_of_observation_segments))

        for i in range(0, self.number_of_observation_segments - 1):
            lower = np.rint(avg_length * (i)).astype(int)
            upper = np.rint(avg_length * (i + 1)).astype(int)
            avg_kappa_1[i] = self.shearable_rod.kappa[0, lower:upper].mean()
        lower = np.rint(avg_length * (self.number_of_observation_segments - 1)).astype(
            int
        )
        avg_kappa_1[-1] = self.shearable_rod.kappa[0, lower:].mean()

        avg_kappa_2 = np.zeros(self.number_of_observation_segments)
        for i in range(0, self.number_of_observation_segments - 1):
            lower = np.rint(avg_length * (i)).astype(int)
            upper = np.rint(avg_length * (i + 1)).astype(int)
            avg_kappa_2[i] = self.shearable_rod.kappa[1, lower:upper].mean()
        lower = np.rint(avg_length * (self.number_of_observation_segments - 1)).astype(
            int
        )
        avg_kappa_2[-1] = self.shearable_rod.kappa[1, lower:].mean()

        state = np.concatenate(
            (
                # rod curvature information
                avg_kappa_1
                * self.base_length
                / (
                    2 * np.pi
                ),  # normalize by the curvature of if the rod is a perfect ring.
                avg_kappa_2
                * self.base_length
                / (
                    2 * np.pi
                ),  # normalize by the curvature of if the rod is a perfect ring.
                # arm tip location
                self.shearable_rod.position_collection[..., -1] / self.base_length,
                # target location
                self.wsol[self.tick] / 1000,  # convert back to meters (~[-1,1])
            )
        )

        return state

    def step(self, action):

        # self.render()

        self.spline_points_func_array_normal_dir[:] = action[
            : self.number_of_control_points
        ]
        self.spline_points_func_array_binormal_dir[:] = action[
            self.number_of_control_points :
        ]

        for _ in range(int(self.num_steps_per_update)):
            self.time_tracker = self.do_step(
                self.StatefulStepper,
                self.stages_and_updates,
                self.simulator,
                self.time_tracker,
                self.sim_dt,
            )
            self.tick += 1

        """ Reward Engineering """
        tip_to_target = (
            self.wsol[self.tick] - self.shearable_rod.position_collection[..., -1]
        ) / 1000
        reward_dist = -np.square(np.linalg.norm(tip_to_target))
        reward = 1.0 * reward_dist  # + reward_orientation

        # observe current state: current as sensed signal
        state = self.get_state()

        """ Done is a boolean to reset the environment when episode is completed """
        done = False

        # Check if the episode blew up and is returning NaNs in the state.
        invalid_values_condition = _isnan_check(state)
        if invalid_values_condition == True:
            reward = -100
            state = np.nan_to_num(self.get_state())
            done = True
            print("Episode blew up after %i steps. Maybe try a smaller dt?" & self.tick)

        if self.tick * self.sim_dt >= self.max_episode_final_time:
            done = True
            # print('Episode has reached max time')

        self._target = self.wsol[self.tick]
        return state, reward, done, {"ctime": self.time_tracker}

    def reset(self):
        self.simulator = BaseSimulator()

        ###--------------ADD ARM TO SIMULATION--------------###
        start = np.zeros((3,))
        direction = np.array([0.0, 1.0, 0.0])  # rod direction: pointing upwards
        normal = np.array([0.0, 0.0, 1.0])
        binormal = np.cross(direction, normal)

        # Set the arm properties after defining rods

        radius_tip = self.radius  # radius of the arm at the tip
        radius_base = radius_tip  # radius of the arm at the base
        radius_along_rod = np.linspace(radius_tip, radius_tip, self.n_elem)

        # Arm is a shearable Cosserat rod
        self.shearable_rod = CosseratRod.straight_rod(
            self.n_elem,
            start,
            direction,
            normal,
            self.base_length,
            base_radius=radius_along_rod,
            density=self.density,
            nu=self.damping,
            youngs_modulus=self.youngs_modulus,
            shear_modulus=self.shear_modulus,
        )

        self.shearable_rod.dissipation_constant_for_torques *= self.torque_damping_ratio

        self.simulator.append(
            self.shearable_rod
        )  # Now rod is ready for simulation, append rod to simulation
        self.simulator.constrain(self.shearable_rod).using(
            OneEndFixedRod, constrained_position_idx=(0,), constrained_director_idx=(0,)
        )
        if self.gravity_mode == 1:
            self.simulator.add_forcing_to(self.shearable_rod).using(
                GravityForces, acc_gravity=np.array([0.0, -9.81 * 1e3, 0.0])
            )  # Gravity points down -- for soft arm, might cause buckling
        elif self.gravity_mode == 2:
            self.simulator.add_forcing_to(self.shearable_rod).using(
                GravityForces, acc_gravity=np.array([0.0, 9.81 * 1e3, 0.0])
            )  # Gravity points up -- arm is hanging upside down.

        # Call back function to collect arm data from simulation
        if self.COLLECT_DATA_FOR_POSTPROCESSING:

            class ArmCallBack(CallBackBaseClass):
                """
                Call back function for Elastica rod
                """

                def __init__(
                    self,
                    step_skip: int,
                    callback_params: dict,
                ):
                    CallBackBaseClass.__init__(self)
                    self.every = step_skip
                    self.callback_params = callback_params

                def make_callback(self, system, time, current_step: int):
                    if current_step % self.every == 0:
                        self.callback_params["time"].append(time)
                        self.callback_params["step"].append(current_step)
                        self.callback_params["position"].append(
                            system.position_collection.copy()
                        )
                        self.callback_params["radius"].append(system.radius.copy())
                        self.callback_params["com"].append(
                            system.compute_position_center_of_mass()
                        )
                        self.callback_params["directors"].append(
                            system.director_collection.copy()
                        )
                        self.callback_params["kappa"].append(system.kappa.copy())
                        self.callback_params["omega_collection"].append(
                            system.omega_collection.copy()
                        )
                        self.callback_params["sigma"].append(system.sigma.copy())
                        self.callback_params["tangents"].append(system.tangents.copy())
                        self.callback_params["velocity_collection"].append(
                            system.velocity_collection.copy()
                        )
                        self.callback_params["acceleration_collection"].append(
                            system.acceleration_collection.copy()
                        )
                        return

            # Collect data using callback function for postprocessing
            self.post_processing_dict_rod = defaultdict(list)
            self.simulator.collect_diagnostics(self.shearable_rod).using(
                ArmCallBack,
                step_skip=self.step_skip,
                callback_params=self.post_processing_dict_rod,
            )

        """
        heuristic scaling of torque - note: this works for a 25:1 slenderness ratio. Bending stiffness scales ~r^4 so if thicker or thinner, 
        the torque_scale value might need to be adjusted. This scaling is more designed for changing the young's modulus 
        """
        self.alpha = self.torque_scale * self.radius * self.youngs_modulus
        # Add muscle torques acting on the arm for actuation
        # MuscleTorquesWithVaryingBetaSplines uses the control points selected by RL to
        # generate torques along the arm.
        self.torque_profile_list_for_muscle_in_normal_dir = defaultdict(list)
        self.spline_points_func_array_normal_dir = []
        self.simulator.add_forcing_to(self.shearable_rod).using(
            MuscleTorquesWithVaryingBetaSplines,
            base_length=self.base_length,
            number_of_control_points=self.number_of_control_points,
            points_func_array=self.spline_points_func_array_normal_dir,
            muscle_torque_scale=self.alpha,
            direction=str("normal"),
            step_skip=self.step_skip,
            max_rate_of_change_of_activation=self.max_rate_of_change_of_activation,
            # max_signal_rate_of_change=4*self.sim_dt,
            torque_profile_recorder=self.torque_profile_list_for_muscle_in_normal_dir,
        )

        self.torque_profile_list_for_muscle_in_binormal_dir = defaultdict(list)
        self.spline_points_func_array_binormal_dir = []
        self.simulator.add_forcing_to(self.shearable_rod).using(
            MuscleTorquesWithVaryingBetaSplines,
            base_length=self.base_length,
            number_of_control_points=self.number_of_control_points,
            points_func_array=self.spline_points_func_array_binormal_dir,
            muscle_torque_scale=self.alpha,
            direction=str("binormal"),
            step_skip=self.step_skip,
            max_rate_of_change_of_activation=self.max_rate_of_change_of_activation,
            # max_signal_rate_of_change=4*self.sim_dt,
            torque_profile_recorder=self.torque_profile_list_for_muscle_in_binormal_dir,
        )

        ###--------------GENERATE TARGET TRAJECTORY--------------###
        """ We are not adding the target into the trajectory to speed things up, you could do so if you wished however"""
        self.tick = 0
        if self.mode == 2:
            # TODO: change this to a object that will update the target position as you go
            target_trajectory = generate_trajectory(
                self.max_episode_final_time, self.sim_dt, self.target_v_scale
            )
            self.wsol = target_trajectory
        elif self.mode == 1:
            end_time = self.max_episode_final_time * 1.1
            numpoints = np.rint(1 / self.sim_dt * end_time).astype(int)
            self.wsol = np.zeros((numpoints, 3))
            self.wsol[:] = self.target_location

        self.sphere_radius = 0.0125 * 1000  # Just for rendering purposes

        ###--------------FINALIZE SIMULATION--------------###
        # Finalize simulation environment. After finalize, you cannot add
        # any forcing, constrain or call back functions
        self.simulator.finalize()

        # do_step, stages_and_updates will be used in step function
        self.do_step, self.stages_and_updates = extend_stepper_interface(
            self.StatefulStepper, self.simulator
        )
        self.time_tracker = np.float64(0.0)

        state = self.get_state()
        self._target = self.wsol[self.tick]

        return state

    def render(self, mode="human", close=False):
        maxwidth = 800
        aspect_ratio = 3 / 4

        if self.viewer is None:
            from gym_softrobot.utils.render import pyglet_rendering

            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)

        if self.renderer is None:
            # Switch renderer depending on configuration
            if RENDERER_CONFIG == RendererType.POVRAY:
                from gym_softrobot.utils.render.povray_renderer import Session
            elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
                from gym_softrobot.utils.render.matplotlib_renderer import Session
            else:
                raise NotImplementedError("Rendering module is not imported properly")
            assert issubclass(
                Session, BaseRenderer
            ), "Rendering module is not properly subclassed"
            assert issubclass(
                Session, BaseElasticaRendererSession
            ), "Rendering module is not properly subclassed"
            self.viewer = pyglet_rendering.SimpleImageViewer(maxwidth=maxwidth)
            self.renderer = Session(width=maxwidth, height=int(maxwidth * aspect_ratio))
            self.renderer.add_rods(
                [self.shearable_rod]
            )  # TODO: maybe need add_rod instead
            self.renderer.add_point(self._target.tolist() + [0], self.sphere_radius)

        # POVRAY
        if RENDERER_CONFIG == RendererType.POVRAY:
            state_image = self.renderer.render(
                maxwidth, int(maxwidth * aspect_ratio * 0.7)
            )
            state_image_side = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.0, -0.5], "look_at", [0.0, 0, 0]),
            )
            state_image_top = self.renderer.render(
                maxwidth // 2,
                int(maxwidth * aspect_ratio * 0.3),
                camera_param=("location", [0.0, 0.3, 0.0], "look_at", [0.0, 0, 0]),
            )

            state_image = np.vstack(
                [state_image, np.hstack([state_image_side, state_image_top])]
            )
        elif RENDERER_CONFIG == RendererType.MATPLOTLIB:
            state_image = self.renderer.render()
        else:
            raise NotImplementedError("Rendering module is not imported properly")

        self.viewer.imshow(state_image)

        return state_image

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None
        if self.renderer:
            self.renderer.close()
            self.renderer = None


if __name__ == "__main__":
    from stable_baselines3 import PPO

    env = SoftArmTracking(gravity_mode=1)
    model = PPO(
        "MlpPolicy",
        env,
        verbose=1,
        tensorboard_log="logs/tensorboard/",
    )
    model.learn(
        total_timesteps=100000,
    )
    model.save(
        "POLICY",
    )

    model = PPO.load("POLICY", env=env)
    obs = env.reset()
    score = 0
    for _ in range(env.num_timesteps_per_episode):
        action, _states = model.predict(obs)
        obs, rewards, dones, info = env.step(action)
        env.render()
        score += rewards
    env.close()

    print("--------------------")
    print("Final score:", np.round(score, 4))
    print("--------------------")



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

from collections import defaultdict

from numba import njit
from tqdm import tqdm

import numpy as np
import pickle

import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.mplot3d import Axes3D

from elastica._linalg import _batch_cross, _batch_matvec, _batch_matrix_transpose
from elastica._calculus import quadrature_kernel
from elastica.external_forces import inplace_addition

from gym_softrobot.utils.actuation.algorithms.algorithm import Algorithm
import gym_softrobot.utils.actuation.objects
# from objects.point import PointTarget
# from objects.cylinder import CylinderConstraint, CylinderTarget
from gym_softrobot.utils.actuation.rod_tools import (
    forward_path,
    backward_path,
    calculate_dilatation,
    sigma_to_shear,
    kappa_to_curvature,
    # calculate_potential_energy
)

from gym_softrobot.utils.actuation.actuations.muscles import TransverseMuscle
from gym_softrobot.utils.actuation.actuations.actuation import _force_induced_couple


@njit(cache=True)
def backward_path_2(dl, director, sigma, kappa, internal_force, internal_couple):

    blocksize = internal_force.shape[1]

    _, voronoi_dilatation = calculate_dilatation(sigma)
    shear = sigma_to_shear(sigma)
    curvature = kappa_to_curvature(kappa, voronoi_dilatation)

    for i in range(blocksize-1):
        internal_force[:, -1-i-1] = internal_force[:, -1-i] + (
            np.cross(curvature[:, -1-i], internal_force[:, -1-i])
            ) * dl
    nu_cross_n = _batch_cross(shear, internal_force)

    for i in range(blocksize-2):
        # maybe kappa need to be devided by voroni_dilatation?
        kappa_cross_m = np.cross(curvature[:, -1-i], internal_couple[:, -1-i])
        internal_couple[:, -1-i-1] = internal_couple[:, -1-i] + (
            kappa_cross_m
            + (nu_cross_n[:, -1-i] + nu_cross_n[:, -1-i-1])/2
            ) * dl
    pass

@njit(cache=True)
def backward_path_3(dl, director, sigma, ns, ms, n1, m1, internal_force, internal_couple):
    shear = np.zeros(sigma.shape)
    # print(shear.shape)
    shear[:, :] = sigma_to_shear(sigma)
    
    n = np.zeros(internal_force.shape)
    n[:, -1] = n1
    for i in range(n.shape[1]-1):
        n[:, -1-i-1] = n[:, -1-i] - (ns[:, -1-i]+ns[:, -1-i-1])*dl[-1-i]/2
    internal_force[:, :] = _batch_matvec(director, n)

    m = np.zeros(internal_force.shape)
    m[:, -1] = m1
    ms[:, :] = -_batch_cross(_batch_matvec(_batch_matrix_transpose(director), shear), n)
    for i in range(m.shape[1]-1):
        m[:, -1-i-1] = m[:, -1-i] - (ms[:, -1-i]+ms[:, -1-i-1])*dl[-1-i]/2
    internal_couple[:, :] = quadrature_kernel(_batch_matvec(director, m))[:, 1:-1]
    return

@njit(cache=True)
def muscle_to_strain(muscle_force, muscle_couple, sigma, kappa, shear_matrix, bend_matrix):

    dilatation, voronoi_dilatation = calculate_dilatation(sigma)

    sigma[0, :] = - muscle_force[0, :] / (shear_matrix[0, 0, :] / dilatation)
    sigma[1, :] = - muscle_force[1, :] / (shear_matrix[1, 1, :] / dilatation)
    sigma[2, :] = - muscle_force[2, :] / (shear_matrix[2, 2, :] / dilatation)

    kappa[0, :] = - muscle_couple[0, :] / (bend_matrix[0, 0, :] / voronoi_dilatation**3)
    kappa[1, :] = - muscle_couple[1, :] / (bend_matrix[1, 1, :] / voronoi_dilatation**3)
    kappa[2, :] = - muscle_couple[2, :] / (bend_matrix[2, 2, :] / voronoi_dilatation**3)
    pass

@njit(cache=True)
def update_muscle_actuation(
    activation, muscle_force, muscle_couple,
    sigma, kappa, internal_force, internal_couple,
    shear_matrix, bend_matrix, eta_activation, pre_activation
    ):

    blocksize = activation.shape[0]

    alpha_bend = np.zeros((3, blocksize))
    alpha = np.zeros((3, blocksize+1))

    for i in range(blocksize):
        for j in range(3):
            alpha_bend[1, i] -= internal_couple[j, i] * muscle_couple[j, i] / bend_matrix[j, j, i]
            alpha_bend[2, i] += (muscle_couple[j, i])**2 / bend_matrix[j, j, i]

    for i in range(blocksize+1):
        for j in range(3):
            alpha[1, i] -= internal_force[j, i] * muscle_force[j, i] / shear_matrix[j, j, i]
            alpha[2, i] += (muscle_force[j, i])**2 / shear_matrix[j, j, i]

    for i in range(blocksize):
        alpha_bend[1, i] += (alpha[1, i] + alpha[1, i+1]) / 2
        alpha_bend[2, i] += (alpha[2, i] + alpha[2, i+1]) / 2

    activation[:] += eta_activation * (-alpha_bend[2, :]*activation + alpha_bend[1, :])

    for i in range(blocksize):
        if activation[i] < 0:
            activation[i] = 0
        elif activation[i] > 1:
            activation[i] = 1
        else:
            pass

    pass

class ForwardBackwardMuscle(Algorithm):
    def __init__(self, rod, muscles, algo_config):
        Algorithm.__init__(self, rod, algo_config)
        self.target = algo_config.get('target', None)
        self.prev_target = None
        self.target = algo_config.get('target', None)
        self.prev_target = None

        self.muscles = muscles
        self.activations = []
        self.pre_activations = []
        for muscle in self.muscles:
            self.activations.append(muscle.activation.copy())
            self.pre_activations.append(muscle.activation.copy())
        self.eta = algo_config.get('eta', 1e-8)

    def update(self, update_weight):

        muscle_forces = np.zeros((3, self.n_elems))
        muscle_couples = np.zeros((3, self.n_elems-1))

        dilatation, _ = calculate_dilatation(self.sigma)
        for muscle, activation in zip(self.muscles, self.activations):
            muscle.set_activation(activation)
            muscle(self.rod, count_flag=False)
            if not (type(muscle) is TransverseMuscle):
                _force_induced_couple(
                    muscle.internal_forces,
                    muscle.muscle_radius_ratio * self.rod.radius / np.sqrt(dilatation),
                    muscle.internal_couples
                )
            inplace_addition(muscle_forces, muscle.internal_forces)
            inplace_addition(muscle_couples, muscle.internal_couples)
        
        # TODO: fix the area and second moment of area change
        muscle_to_strain(
            muscle_forces, muscle_couples,
            self.sigma, self.kappa,
            self.shear_matrix, self.bend_matrix
        )
        
        forward_path(self.dl, self.sigma, self.kappa, self.position, self.director)

        # if type(self.target) is PointTarget:
        #     terminal_cost_partial_end_position, terminal_cost_partial_end_director = (
        #         self.target.terminal_cost_partial_end_state(
        #             self.position[:, -1],
        #             end_director=self.director[:, :, -1]
        #         )
        #     )
        # else:
        #     pass

        # self.internal_force[:, -1] = - self.director[:, :, -1] @ terminal_cost_partial_end_position
        # if not (terminal_cost_partial_end_director == None):
        #     self.internal_couple[:, -1] = terminal_cost_partial_end_director


        midpoint_position = (self.position[:, :-1] + self.position[:, 1:]) / 2
        ns = np.zeros(midpoint_position.shape)
        ms = np.zeros(midpoint_position.shape)
        n1 = np.zeros(3)
        m1 = np.zeros(3)

        # for constraint in self.constraints:
        #     constraint.calculate_cost(midpoint_position, self.radii)
        #     ns[:, :] -= constraint.running_cost_partial_position(midpoint_position)

        # self.target.calculate_cost(self.position[:, -1], midpoint_position, self.radius)
        terminal_cost_partial_end_position, terminal_cost_partial_end_director = (
                self.target.terminal_cost_partial_end_state(
                    self.position[:, -1],
                    end_director=self.director[:, :, -1]
                )
            )
        n1[:] = - terminal_cost_partial_end_position

        if objects.object.Constraint in self.target.__class__.__mro__:
            self.target.running_cost(midpoint_position, self.radius)
            ns[:, :] -= self.target.running_cost_partial_position(midpoint_position)
        # print(self.dl)
        # print(ns.shape, self.sigma.shape)

        backward_path_3(self.dl, self.director, self.sigma, ns, ms, n1, m1, self.internal_force, self.internal_couple)

        # backward_path(self.dl, self.director, self.sigma, ns, ms, n1, m1, self.internal_force, self.internal_couple)

        dilatation, _ = calculate_dilatation(self.sigma)
        for i, muscle in enumerate(self.muscles):
            muscle.set_activation(np.ones(self.n_elems-1))
            muscle(self.rod, count_flag=False)
            if not (type(muscle) is TransverseMuscle):
                _force_induced_couple(
                    muscle.internal_forces,
                    muscle.muscle_radius_ratio * self.rod.radius / np.sqrt(dilatation),
                    muscle.internal_couples
                )

            update_muscle_actuation(
                self.activations[i],
                muscle.internal_forces, muscle.internal_couples,
                self.sigma, self.kappa,
                self.internal_force, self.internal_couple,
                self.shear_matrix, self.bend_matrix,
                self.eta, #max(1e-10/update_weight, 1e-15)
                self.pre_activations[i]
            )

        return update_weight+1

    def run(self, continuum_target_cost_weight=None, plot_flag=False, iter_number=100000):
        print("Running algorithm with target:", self.target)
        update_weight = 1
        for _ in tqdm(range(iter_number)):
            update_weight = self.update(update_weight)
        return



"""
Created on Apr. 27, 2021
@author: Heng-Sheng (Hanson) Chang
"""

from __future__ import division
from __future__ import print_function

from collections import defaultdict

from numba import njit
from numpy.core.shape_base import block
from tqdm import tqdm

import numpy as np
import pickle

import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib.colors as mcolors

from elastica._linalg import _batch_cross, _batch_matvec, _batch_matrix_transpose
from elastica._calculus import quadrature_kernel
from elastica._rotations import _rotate, _inv_rotate
from elastica.external_forces import inplace_addition

from gym_softrobot.utils.actuation.algorithms.algorithm import Algorithm
from gym_softrobot.utils.actuation.rod_tools import (
    forward_path,
    calculate_dilatation,
    sigma_to_shear,
    kappa_to_curvature,
    _lab_to_material,
    _material_to_lab,
    _batch_cross,
    # _trapezoidal,
    average
)

class ForwardBackwardSmooth(Algorithm):
    def __init__(self, rod, algo_config, data):
        Algorithm.__init__(self, rod, algo_config)

        self.argument_rotational = np.zeros((3, rod.n_elems-1))
        self.argument_translational = np.zeros((3, rod.n_elems))
        self.costate_rotational = np.zeros((3, rod.n_elems-1))
        self.costate_translational = np.zeros((3, rod.n_elems))

        self.data = data
        
        self.s_position_jump_index = np.full(rod.n_elems+1, -1, dtype=int)
        self.s_director_jump_index = np.full(rod.n_elems, -1, dtype=int)
        for index, s in enumerate(data.s_position):
            for s_index in range(self.s_position.shape[0]):
                if s_index == 0:
                    s_range_plus = self.s_position[1]/2
                    s_range_minus = 0
                elif s_index == self.s_position.shape[0]-1:
                    s_range_plus = 0
                    s_range_minus = (self.s_position[-1]-self.s_position[-2])/2
                else:
                    s_range_plus = (self.s_position[s_index+1]-self.s_position[s_index])/2
                    s_range_minus = (self.s_position[s_index]-self.s_position[s_index-1])/2
                s_plus = self.s_position[s_index] + s_range_plus
                s_minus = self.s_position[s_index] - s_range_minus
                
                if (s_plus-s)*(s-s_minus) >= -1e-5:
                    self.s_position_jump_index[s_index] = index
                    break        
        self.director_flag = data.director_flag
        if self.director_flag:
            for index, s in enumerate(data.s_director):
                for s_index in range(self.s_director.shape[0]):
                    if s_index == 0:
                        s_range_plus = self.s_director[1]/2
                        s_range_minus = 0
                    elif s_index == self.s_director.shape[0]-1:
                        s_range_plus = 0
                        s_range_minus = (self.s_director[-1]-self.s_director[-2])/2
                    else:
                        s_range_plus = (self.s_director[s_index+1]-self.s_director[s_index])/2
                        s_range_minus = (self.s_director[s_index]-self.s_director[s_index-1])/2
                    s_plus = self.s_director[s_index] + s_range_plus
                    s_minus = self.s_director[s_index] - s_range_minus
                    
                    if (s_plus-s)*(s-s_minus) >= -1e-5:
                        self.s_director_jump_index[s_index] = index
                        break
            self.update = self.update_with_director_jump
        else:
            self.update = self.update_without_director_jump

    def calculate_cost(self,):
        cost = 0
        cost += self.energy_cost(
            self.dl, self.shear, self.kappa, self.rest_shear, self.rest_kappa,
            self.shear_matrix, self.bend_matrix,
        )
        cost += self.argument_cost(
            self.dl, self.argument_rotational, self.argument_translational, self.config.argument_weight
        )
        cost += self.data_position_cost(
            self.position,
            self.s_position_jump_index,
            self.data.noisy_position,
            self.config.data_deviation_weight_cost_position
        )
        if self.director_flag:
            cost += self.data_director_cost(
                self.director,
                self.s_director_jump_index,
                self.data.noisy_director,
                self.config.data_deviation_weight_cost_director
            )

        return cost

    @staticmethod
    @njit(cache=True)
    def energy_cost(
        dl, shear, kappa, rest_shear, rest_kappa,
        shear_matrix, bend_matrix
    ):
        cost = 0
        for n in range(dl.shape[0]):
            for i in range(3):
                cost += (shear[i, n]-rest_shear[i, n])**2 * shear_matrix[i, i, n] * dl[i]
        voronoi_dl = (dl[:-1] + dl[1:]) / 2
        for n in range(voronoi_dl.shape[0]):
            for i in range(3):
                cost += (kappa[i, n]-rest_kappa[i, n])**2 * bend_matrix[i, i, n] * voronoi_dl[i]
        return 0.5*cost

    @staticmethod
    @njit(cache=True)
    def argument_cost(
        dl, argument_rotational, argument_translational, argument_weight
    ):
        cost = 0
        for n in range(argument_translational.shape[1]):
            for i in range(argument_translational.shape[0]):
                cost += argument_translational[i, n] ** 2 * dl[n]
        voronoi_dl = (dl[:-1] + dl[1:]) / 2
        for n in range(argument_rotational.shape[1]):
            for i in range(argument_rotational.shape[0]):
                cost += argument_rotational[i, n] ** 2 * voronoi_dl[n]
        return 0.5*cost*argument_weight

    @staticmethod
    @njit(cache=True)
    def data_position_cost(
        position,
        s_position_jump_index,
        data_position,
        data_deviation_weight_cost_position,
    ):
        cost = 0
        for n, index in enumerate(s_position_jump_index):
            if index != -1:
                cost += data_deviation_weight_cost_position * position_cost(position[:, n], data_position[:, index])
        return cost

    @staticmethod
    @njit(cache=True)
    def data_director_cost(
        director,
        s_director_jump_index,
        data_director,
        data_deviation_weight_cost_director
    ):
        cost = 0
        for n, index in enumerate(s_director_jump_index):
            if index != -1:
                cost += data_deviation_weight_cost_director * director_cost(director[:, :, n], data_director[:, :, index])
        return cost

    def update_strain_and_forward_path(self,):
        self.update_argument(
            self.argument_translational, self.argument_rotational,
            self.costate_translational, self.costate_rotational,
            self.config.argument_weight, self.config.step_size
        )
        self.forward_path_strain(
            self.dl, self.shear, self.kappa,
            self.argument_rotational, self.argument_translational,
        )
        forward_path(
            self.dl, self.shear, self.kappa, 
            self.position, self.director,
        )

    def update_without_director_jump(self, ):
        self.backward_path_without_director_jump(
            self.dl, self.position, self.director, self.shear, self.kappa,
            self.internal_force, self.internal_couple,
            self.s_position_jump_index, self.data.noisy_position,
            self.config.data_deviation_weight_cost_position,
        )
        self.costate_backward_path(
            self.dl, self.kappa, self.shear,
            self.rest_kappa, self.rest_shear,
            self.bend_matrix, self.shear_matrix,
            self.costate_rotational, self.costate_translational,
            self.internal_couple, self.internal_force,
        )
        self.update_strain_and_forward_path()

    def update_with_director_jump(self, ):
        self.backward_path_with_director_jump(
            self.dl, self.position, self.director, self.shear, self.kappa,
            self.internal_force, self.internal_couple,
            self.s_position_jump_index, self.data.noisy_position,
            self.config.data_deviation_weight_cost_position,
            self.s_director_jump_index, self.data.noisy_director,
            self.config.data_deviation_weight_cost_director
        )
        self.costate_backward_path(
            self.dl, self.kappa, self.shear,
            self.rest_kappa, self.rest_shear,
            self.bend_matrix, self.shear_matrix,
            self.costate_rotational, self.costate_translational,
            self.internal_couple, self.internal_force,
        )
        self.update_strain_and_forward_path()

    def run(self, iter_number=1_000_000):
        print("Running forward-backward algorithm to smooth the data")
        cost = np.zeros(iter_number)
        for k in tqdm(range(iter_number)):
            self.update()
            cost[k] = self.calculate_cost()
        return cost

    @staticmethod
    @njit(cache=True)
    def forward_path_strain(
        dl, shear, kappa,
        argument_rotational, argument_translational,
    ):
        blocksize = dl.shape[0]
        for n in range(blocksize-2):
            kappa[:, n+1] = kappa[:, n] + argument_rotational[:, n]*dl[n]
            shear[:, n+1] = shear[:, n] + argument_translational[:, n]*dl[n]
        shear[:, -1] = shear[:, -2] + argument_translational[:, -2]*dl[-2]

    @staticmethod
    @njit(cache=True)
    def update_argument(
        argument_translational, argument_rotational,
        costate_translational, costate_rotational,
        argument_weight, step_size
    ):
        for n in range(argument_rotational.shape[1]):
            argument_rotational[:, n] += (
                step_size * (costate_rotational[:, n] - argument_weight*argument_rotational[:, n])
            )
            argument_translational[:, n] += (
                step_size * (costate_translational[:, n] - argument_weight*argument_translational[:, n])
            )
        argument_translational[:, -1] += (
            step_size * (costate_translational[:, -1] - argument_weight*argument_translational[:, -1])
        )

    @staticmethod
    @njit(cache=True)
    def costate_backward_path(
        dl, kappa, shear,
        rest_kappa, rest_shear,
        bend_matrix, shear_matrix,
        costate_rotational, costate_translational,
        internal_couple, internal_force,
    ):
        blocksize = dl.shape[0]
        for i in range(3):
            costate_rotational[i, -1] = - (
                -internal_couple[i, -1] + bend_matrix[i, i, -1]*(kappa[i, -1]-rest_kappa[i, -1])
            )*dl[-1]
            costate_translational[i, -1] = - (
                -internal_force[i, -1] + shear_matrix[i, i, -1]*(shear[i, -1]-rest_shear[i, -1])
            )*dl[-1]

        for n in range(blocksize-2):
            for i in range(3):
                costate_rotational[i, -1-n-1] = costate_rotational[i, -1-n] - (
                    -internal_couple[i, -1-n-1] + bend_matrix[i, i, -1-n-1]*(kappa[i, -1-n-1]-rest_kappa[i, -1-n-1])
                )*dl[-1-n-1]
                costate_translational[i, -1-n-1] = costate_translational[i, -1-n] - (
                    -internal_force[i, -1-n-1] + shear_matrix[i, i, -1-n-1]*(shear[i, -1-n-1]-rest_shear[i, -1-n-1])
                )*dl[-1-n-1]
        
        for i in range(3):
            costate_translational[i, 0] = costate_translational[i, 1] - (
                -internal_force[i, 0] + shear_matrix[i, i, 0]*(shear[i, 0]-rest_shear[i, 0])
            )*dl[0]
        pass

    @staticmethod
    @njit(cache=True)
    def backward_path_without_director_jump(
        dl, position, director, shear, kappa,
        internal_force, internal_couple,
        s_position_jump_index, data_position, data_deviation_weight_cost_position,
    ):
        # calculate internal force
        internal_force_lab_frame = (
            calculate_internal_force_with_jumping_condition(
                position, director, internal_force,
                s_position_jump_index, data_position, data_deviation_weight_cost_position,
            )
        )

        # calculate internal couple
        internal_couple_lab_frame = (
            calculate_internal_couple_without_jumping_condition(
                dl, director, shear, internal_force_lab_frame, internal_couple,
            )
        )
    
    @staticmethod
    @njit(cache=True)
    def backward_path_with_director_jump(
        dl, position, director, shear, kappa,
        internal_force, internal_couple,
        s_position_jump_index, data_position, data_deviation_weight_cost_position,
        s_director_jump_index, data_director, data_deviation_weight_cost_director,
    ):
        # calculate internal force
        internal_force_lab_frame = (
            calculate_internal_force_with_jumping_condition(
                position, director, internal_force,
                s_position_jump_index, data_position, data_deviation_weight_cost_position,
            )
        )

        # calculate internal couple
        internal_couple_lab_frame = (
            calculate_internal_couple_with_jumping_condition(
                dl, director, shear, internal_force_lab_frame, internal_couple,
                s_director_jump_index, data_director, data_deviation_weight_cost_director,
            )
        )

@njit(cache=True)
def calculate_internal_force_with_jumping_condition(
    position, director, internal_force,
    s_position_jump_index, data_position, data_deviation_weight_cost_position,
):
    blocksize = internal_force.shape[1]
    internal_force_lab_frame = np.zeros((3, blocksize))
    # print(internal_force_lab_frame[:,0])
    # calculate internal force in lab frame
    for i in range(blocksize-1):
        # this is the jumping condition
        if s_position_jump_index[-1-i] != -1:
            internal_force_lab_frame[:, -1-i] -= data_deviation_weight_cost_position * (
                position[:, -1-i] - data_position[:, s_position_jump_index[-1-i]]
            )
        internal_force_lab_frame[:, -1-i-1] = internal_force_lab_frame[:, -1-i].copy()

    internal_force[:, :] = _lab_to_material(director, internal_force_lab_frame)
    return internal_force_lab_frame

@njit(cache=True)
def calculate_internal_couple_with_jumping_condition(
    dl, director, shear, internal_force_lab_frame, internal_couple,
    s_director_jump_index, data_director, data_deviation_weight_cost_director,
):
    blocksize = internal_couple.shape[1]+1
    internal_couple_lab_frame = np.zeros((3, blocksize))

    ms = - _batch_cross(_material_to_lab(director, shear), internal_force_lab_frame)
    for i in range(blocksize-1):
        # this is the jumping condition
        if s_director_jump_index[-1-i] != -1:
            internal_couple_lab_frame[:, -1-i] += data_deviation_weight_cost_director * (
                difference_between_directors(
                    director[:, :, -1-i], data_director[:, :, s_director_jump_index[-1-i]]
                )
            )

        internal_couple_lab_frame[:, -1-i-1] = (
            internal_couple_lab_frame[:, -1-i] - (ms[:, -1-i]+ms[:, -1-i-1])*dl[-1-i]/2
        )

    internal_couple[:, :] = couple_in_lab_transform_to_material(
        internal_couple_lab_frame, director
    )

    return internal_couple_lab_frame

@njit(cache=True)
def calculate_internal_couple_without_jumping_condition(
    dl, director, shear, internal_force_lab_frame, internal_couple,
):
    blocksize = internal_couple.shape[1]+1
    internal_couple_lab_frame = np.zeros((3, blocksize))

    ms = - _batch_cross(_material_to_lab(director, shear), internal_force_lab_frame)
    for i in range(blocksize-1):
        internal_couple_lab_frame[:, -1-i-1] = (
            internal_couple_lab_frame[:, -1-i] - (ms[:, -1-i]+ms[:, -1-i-1])*dl[-1-i]/2
        )

    internal_couple[:, :] = couple_in_lab_transform_to_material(
        internal_couple_lab_frame, director
    )
    return internal_couple_lab_frame

@njit(cache=True)
def difference_between_directors(director1, director2):
    difference = _mat1mat2t(director1, director2) - _mat1mat2t(director2, director1)
    return _matTvec(director1, _vec(difference))

@njit(cache=True)
def couple_in_lab_transform_to_material(internal_couple_lab_frame, director):
    blocksize = internal_couple_lab_frame.shape[1]
    internal_couple = np.zeros((3, blocksize-1))
    for i in range(blocksize-1):
        middle_couple = (internal_couple_lab_frame[:, i] + internal_couple_lab_frame[:, i+1]) / 2
        middle_director = calculate_middle_director(director[:, :, i:i+2])
        internal_couple[:, i] = _matvec(middle_director, middle_couple)
    return internal_couple

@njit(cache=True)
def calculate_middle_director(directors):
    vector = _inv_rotate(directors)
    return _rotate(directors[:, :, 0:1], 0.5, vector)[:, :, 0]

@njit(cache=True)
def _matvec(matrix, vec):
    result = np.zeros(3,)
    for i in range(3):
        for j in range(3):
            result[i] += matrix[i][j] * vec[j]
    return result

@njit(cache=True)
def _matTvec(matrix, vec):
    result = np.zeros(3,)
    for i in range(3):
        for j in range(3):
            result[i] += matrix[j][i] * vec[j]
    return result

@njit(cache=True)
def _mat1mat2t(matrix1, matrix2):
    result = np.zeros((3, 3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                result[i, j] += matrix1[i][k] * matrix2[j][k]
    return result

@njit(cache=True)
def _vec(skew_matrix):
    return np.array([
        -skew_matrix[1, 2],
        skew_matrix[0, 2],
        -skew_matrix[0,1]
    ])

@njit(cache=True)
def position_cost(position1, position2):
    cost = 0
    for i in range(3):
        cost += (position1[i]-position2[i])**2
    cost = 0.5*np.sqrt(cost)
    return cost

@njit(cache=True)
def director_cost(director1, director2):
    identity = np.eye(3)
    matrix = identity - _mat1mat2t(director1, director2)
    cost = np.trace(_mat1mat2t(matrix, matrix))
    return cost



"""
Created on Oct. 19, 2020
@author: Heng-Sheng (Hanson) Chang
"""

import numpy as np

from gym_softrobot.utils.actuation.objects.object import Target

from elastica._rotations import _inv_rotate

class PointTarget(Target):
    def __init__(self, position, director=None, target_cost_weight=None):
        Target.__init__(self, 'point', position, director, target_cost_weight)
    
    def __str__(self,):
        result = "Target is a point with position at (%f, %f, %f)." % (self.position[0], self.position[1], self.position[2])
        if not (self.director == None):
            result += "and with target orientation."
        return result

    def terminal_cost(self, end_position, end_director=None):
        self.cost[0] = 0.5 * self.target_cost_weight * np.linalg.norm(self.position - end_position)**2
        if self.director is not None:
            director_collection = np.zeros((3, 3, 2))
            director_collection[:, :, 0] = end_director
            director_collection[:, :, 1] = self.director
            # print(director_collection)
            rotate_axis_and_angle = _inv_rotate(director_collection)
            # print(rotate_axis_and_angle)
            # quit()

        return self.cost[0]

    def terminal_cost_partial_end_state(self, end_position, end_director=None):
        terminal_cost_partial_end_position = self.target_cost_weight * (end_position - self.position)

        if self.director is not None:
            director_collection = np.zeros((3, 3, 2))
            director_collection[:, :, 0] = end_director
            director_collection[:, :, 1] = self.director
            rotate_axis_and_angle = _inv_rotate(director_collection)
            terminal_cost_partial_end_director = self.target_cost_weight * rotate_axis_and_angle[:, 0]
            return terminal_cost_partial_end_position, terminal_cost_partial_end_director

        return terminal_cost_partial_end_position, None

    def calculate_cost(self, position, director, radii):
        return self.terminal_cost(position[:, -1], director[:, :, -1])



from elastica.callback_functions import CallBackBaseClass


class RigidCylinderCallBack(CallBackBaseClass):
    """
    Call back function for continuum snake
    """

    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["time"].append(time)
            self.callback_params["step"].append(current_step)
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params["velocity"].append(system.velocity_collection.copy())

        return

class RodCallBack(CallBackBaseClass):
    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["time"].append(time)
            self.callback_params["radius"].append(system.radius.copy())
            self.callback_params["dilatation"].append(system.dilatation.copy())
            self.callback_params["voronoi_dilatation"].append(system.voronoi_dilatation.copy())
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params["director"].append(system.director_collection.copy())
            self.callback_params["velocity"].append(system.velocity_collection.copy())
            self.callback_params["omega"].append(system.omega_collection.copy())
            self.callback_params["sigma"].append(system.sigma.copy())
            self.callback_params["kappa"].append(system.kappa.copy())
        return

class ExternalLoadCallBack(CallBackBaseClass):
    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params['external_force'].append(system.external_forces.copy())
            self.callback_params['external_couple'].append(system.external_torques.copy())
        return

class CylinderCallBack(CallBackBaseClass):
    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["time"].append(time)
            self.callback_params["radius"].append(system.radius)
            self.callback_params["height"].append(system.length)
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params['director'].append(system.director_collection.copy())
        return

class SphereCallBack(CallBackBaseClass):
    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["time"].append(time)
            self.callback_params["radius"].append(system.radius)
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params["director"].append(system.director_collection.copy())
        return

class AlgorithmCallBack(CallBackBaseClass):
    def __init__(self, step_skip: int, callback_params: dict):
        CallBackBaseClass.__init__(self)
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, system, time, current_step: int):
        if current_step % self.every == 0:
            self.callback_params["position"].append(system.position_collection.copy())
            self.callback_params["director"].append(system.director_collection.copy())
            self.callback_params["sigma"].append(system.sigma.copy())
            self.callback_params["kappa"].append(system.kappa.copy())
        return

class OthersCallBack:
    def __init__(self, step_skip: int, callback_params: dict):
        self.current_step = 0
        self.every = step_skip
        self.callback_params = callback_params

    def make_callback(self, time, **kwargs):
        if self.current_step % self.every == 0:
            self.callback_func(time, **kwargs)
        self.current_step += 1

    def callback_func(self, time, **kwargs):
        self.callback_params['time'].append(time)
        for key, value in kwargs.items():
            # make sure value is a numpy array
            self.callback_params[key].append(value.copy())

    def save_data(self, **kwargs):

        import pickle

        print("Saving additional data to simulation_others.pickle file...", end='\r')

        with open("simulation_others.pickle", "wb") as others_file:
            data = dict(
                time_series_data=self.callback_params,
                **kwargs
            )
            pickle.dump(data, others_file)

        print("Saving additional data to simulation_others.pickle file... Done!")



from dataclasses import dataclass
import numpy as np
from numba import njit
from elastica._rotations import _rotate

from elastica.boundary_conditions import FreeRod


class ControllableFixConstraint(FreeRod):
    @dataclass
    class _Controller:
        index: int
        flag: bool = True
        reduction_ratio: float = 1.0

        def __bool__(self):
            return self.flag

        def turn_on(self):
            self.flag = True

        def turn_off(self):
            self.flag = False

    """ Modelled after sucker on octopus arm"""

    def __init__(self, index, reduction_ratio=1.0, **kwargs):
        super().__init__(**kwargs)
        self.controller = self._Controller(index=index, reduction_ratio=reduction_ratio)

    @property
    def get_controller(self):
        return self.controller

    def constrain_values(self, rod, time):
        return
        # if self.controller.flag:
        #    self.nb_compute_constrain_values(
        #        rod.position_collection,
        #        rod.director_collection,
        #        self.controller.index
        #    )

    def constrain_rates(self, rod, time):
        if self.controller.flag:
            self.nb_compute_constrain_rates(
                rod.velocity_collection,
                rod.omega_collection,
                self.controller.index,
                self.controller.reduction_ratio,
            )

    # @staticmethod
    # @njit(cache=True)
    # def nb_compute_constrain_values(
    #    position_collection, director_collection, index
    # ):
    #    position_collection[2, index] = 0
    #    director_collection[1, :, index] = np.array([0.0,0.0,1.0])

    @staticmethod
    @njit(cache=True)
    def nb_compute_constrain_rates(
        velocity_collection, omega_collection, index, reduction_ratio
    ):
        velocity_collection[..., index] *= 1.0 - reduction_ratio
        omega_collection[..., index] *= 1.0 - reduction_ratio



__doc__ = """
Module contains elastica interface to create octopus model.
This model specifically create 8-arm octopus with each arm controlled by muscle model.
Suposedly, this model represend simplified control of arm coordination.
"""

from typing import Optional, Tuple
import json

from copy import deepcopy

import numpy as np

from elastica import *
from elastica.timestepper import extend_stepper_interface
from elastica.experimental.interaction import AnisotropicFrictionalPlaneRigidBody

from gym_softrobot.utils.custom_elastica.joint import FixedJoint2Rigid
from gym_softrobot.utils.custom_elastica.constraint import BodyBoundaryCondition

# from gym_softrobot.utils.actuation.forces.drag_force import DragForce
from gym_softrobot.utils.actuation.actuations.muscles.longitudinal_muscle import (
    LongitudinalMuscle,
)
from gym_softrobot.utils.actuation.actuations.muscles.transverse_muscle import (
    TransverseMuscle,
)
from gym_softrobot.utils.actuation.actuations.muscles.muscle import ApplyMuscle

from gym_softrobot.envs.octopus.build import create_es_muscle_layers

from scipy.spatial.transform import Rotation as Rot

# default parameters
ARM_MATERIAL = {
    # Arm properties
    "density": 1000.0,
    "youngs_modulus": 1.5e4,
    "shear_modulus": 1.5e4 / (1.0 + 0.5),  # 0.5 Poisson Ratio
    "nu": 0.20,
    "nu_scale": 1e-2,
}
DEFAULT_SCALE_LENGTH = {
    # Arm length scale
    "base_length": 0.25,
    "base_radius": 0.013,
    "tip_radius": 0.0042,  # https://www.liebertpub.com/doi/10.1089/soro.2019.0082
    "head_radius": 0.04,
}
HEAD_PROPERTIES = {
    # Head length scale
    "head_density": 50.0,
    # Head properties
    "body_arm_k": 1e6,
    "body_arm_kt": 1e2,
    "body_arm_nu": 1e-3,
}


def build_arm(
    n_elem: int, start: np.ndarray, direction: np.ndarray, normal: np.ndarray
) -> CosseratRod:
    """Set up properties"""
    arm_material = deepcopy(ARM_MATERIAL)
    arm_material["n_elements"] = n_elem
    arm_material["start"] = start
    arm_material["direction"] = direction
    arm_material["normal"] = normal
    arm_material["base_length"] = DEFAULT_SCALE_LENGTH["base_length"]
    arm_material["base_radius"] = np.linspace(
        DEFAULT_SCALE_LENGTH["base_radius"], DEFAULT_SCALE_LENGTH["tip_radius"], n_elem
    )
    arm_material["nu"] *= (
        (arm_material["base_radius"] / DEFAULT_SCALE_LENGTH["base_radius"]) ** 2.0
    ) * arm_material["nu_scale"]

    rod = CosseratRod.straight_rod(**arm_material)
    return rod


def build_octopus_muscles(simulator, n_elem: int = 40):
    """
    build_octopus_muscles

    Parameters
    ----------
    simulator : PyElastica Simulator
    n_elem : int
    """

    """ Configuration"""
    n_arm = 8

    """ Set up an arm """
    L0 = DEFAULT_SCALE_LENGTH["base_length"]
    r0 = DEFAULT_SCALE_LENGTH["base_radius"]
    head_radius = DEFAULT_SCALE_LENGTH["head_radius"]

    # Levi (2015) - Figure 3B
    shearable_rods = []  # arms
    angles_offset = 45.0 / 2
    angles = [angles_offset + 45 * arm_i for arm_i in range(n_arm)]
    for angle in angles:
        # Rotate director
        rot = Rot.from_euler("z", angle, degrees=True)
        arm_pos = rot.apply([head_radius, 0.0, 0.0])
        arm_dir = rot.apply([1.0, 0.0, 0.0])
        arm_normal = np.array([0.0, 0.0, 1.0])
        # Build Arm
        rod = build_arm(n_elem, arm_pos, arm_dir, arm_normal)
        shearable_rods.append(rod)
        simulator.append(rod)

    """ Add head """
    start = np.zeros((3,))
    start[2] = -r0 * 2
    direction = np.array([0.0, 0.0, 1.0])
    normal = np.array([0.0, 1.0, 0.0])
    density = HEAD_PROPERTIES["head_density"]
    rigid_rod = Cylinder(start, direction, normal, r0 * 2, head_radius, density)
    simulator.append(rigid_rod)

    """ Constraint body """
    simulator.constrain(rigid_rod).using(
        # Upright rigid rod need restoration force/torque against the floor
        BodyBoundaryCondition,
        constrained_position_idx=(0,),
        constrained_director_idx=(0,),
    )

    """ Set up boundary conditions """
    _k = HEAD_PROPERTIES["body_arm_k"]
    _kt = HEAD_PROPERTIES["body_arm_kt"]
    _nu = HEAD_PROPERTIES["body_arm_nu"]
    for arm_i in range(n_arm):
        simulator.connect(
            first_rod=rigid_rod,
            second_rod=shearable_rods[arm_i],
            first_connect_idx=-1,
            second_connect_idx=0,
        ).using(
            FixedJoint2Rigid,
            k=_k,
            nu=_nu,
            kt=_kt,
            angle=angles[arm_i],
            radius=head_radius,
        )

    """ Add drag force """
    """
    dl = base_length_of_arm / n_elem # TODO: base length of arm
    fluid_factor = 1
    r_bar = (radius_base + radius_tip) / 2
    sea_water_dentsity = 1022
    c_per = 0.41 / sea_water_dentsity / r_bar / dl * fluid_factor
    c_tan = 0.033 / sea_water_dentsity / np.pi / r_bar / dl * fluid_factor

    simulator.add_forcing_to(self.shearable_rod).using(
        DragForce,
        rho_environment=sea_water_dentsity,
        c_per=c_per,
        c_tan=c_tan,
        system=self.shearable_rod,
        step_skip=self.step_skip,
        callback_params=self.rod_parameters_dict
    )
    """

    """ Add muscle actuation """
    tm_activations = []
    for i in range(n_arm):
        muscle_layers = create_es_muscle_layers(
            shearable_rods[i].radius, DEFAULT_SCALE_LENGTH["base_radius"]
        )
        simulator.add_forcing_to(shearable_rods[i]).using(
            ApplyMuscle,
            muscles=muscle_layers,
            step_skip=10000,  # Not relavent
            callback_params_list=[],
        )
        tm_activations.append(muscle_layers)  # 2 for TM

    return shearable_rods, rigid_rod, tm_activations


def build_two_arms(simulator, n_elem: int = 40):
    """
    build_two_arms

    Parameters
    ----------
    simulator : PyElastica Simulator
    n_elem : int
    """

    """ Configuration"""
    n_arm = 2

    """ Set up an arm """
    L0 = DEFAULT_SCALE_LENGTH["base_length"]
    r0 = DEFAULT_SCALE_LENGTH["base_radius"]
    head_radius = DEFAULT_SCALE_LENGTH["head_radius"]

    # Levi (2015) - Figure 3B
    shearable_rods = []  # arms
    angles_offset = 90.0
    angles = [angles_offset + 180.0 * arm_i for arm_i in range(n_arm)]
    for angle in angles:
        # Rotate director
        rot = Rot.from_euler("z", angle, degrees=True)
        arm_pos = rot.apply([head_radius, 0.0, 0.0])
        arm_dir = rot.apply([1.0, 0.0, 0.0])
        arm_normal = np.array([0.0, 0.0, 1.0])
        # Build Arm
        rod = build_arm(n_elem, arm_pos, arm_dir, arm_normal)
        shearable_rods.append(rod)
        simulator.append(rod)

    """ Add head """
    start = np.zeros((3,))
    start[2] = -r0 * 2
    direction = np.array([0.0, 0.0, 1.0])
    normal = np.array([0.0, 1.0, 0.0])
    density = HEAD_PROPERTIES["head_density"]
    rigid_rod = Cylinder(start, direction, normal, r0 * 2, head_radius, density)
    simulator.append(rigid_rod)

    """ Constraint body """
    simulator.constrain(rigid_rod).using(
        # Upright rigid rod need restoration force/torque against the floor
        BodyBoundaryCondition,
        constrained_position_idx=(0,),
        constrained_director_idx=(0,),
    )

    """ Set up boundary conditions """
    _k = HEAD_PROPERTIES["body_arm_k"]
    _kt = HEAD_PROPERTIES["body_arm_kt"]
    _nu = HEAD_PROPERTIES["body_arm_nu"]
    for arm_i in range(n_arm):
        simulator.connect(
            first_rod=rigid_rod,
            second_rod=shearable_rods[arm_i],
            first_connect_idx=-1,
            second_connect_idx=0,
        ).using(
            FixedJoint2Rigid,
            k=_k,
            nu=_nu,
            kt=_kt,
            angle=angles[arm_i],
            radius=head_radius,
        )

    """ Add drag force """
    """
    dl = base_length_of_arm / n_elem # TODO: base length of arm
    fluid_factor = 1
    r_bar = (radius_base + radius_tip) / 2
    sea_water_dentsity = 1022
    c_per = 0.41 / sea_water_dentsity / r_bar / dl * fluid_factor
    c_tan = 0.033 / sea_water_dentsity / np.pi / r_bar / dl * fluid_factor

    simulator.add_forcing_to(self.shearable_rod).using(
        DragForce,
        rho_environment=sea_water_dentsity,
        c_per=c_per,
        c_tan=c_tan,
        system=self.shearable_rod,
        step_skip=self.step_skip,
        callback_params=self.rod_parameters_dict
    )
    """

    """ Add muscle actuation """
    arm_activations = []
    for i in range(n_arm):
        muscle_layers = create_es_muscle_layers(
            shearable_rods[i].radius, DEFAULT_SCALE_LENGTH["base_radius"]
        )
        simulator.add_forcing_to(shearable_rods[i]).using(
            ApplyMuscle,
            muscles=muscle_layers,
            step_skip=10000,  # Not relavent
            callback_params_list=[],
        )
        arm_activations.append(muscle_layers)  # 2 for TM

    return shearable_rods, rigid_rod, arm_activations



"""
Created on Sep. 23, 2021
@author: Heng-Sheng (Hanson) Chang
"""

import os, shutil
import matplotlib.pyplot as plt

class FrameBase(object):
    def __init__(self, figure_name, folder_name=None, fig_dict=None):
        self.figure_name = figure_name
        self.folder_name = folder_name
        self.frame_count = 0
        self.fig_dict = fig_dict
        self.check_folder()

    def check_folder(self,):
        if not (self.folder_name is None):
            if os.path.exists(self.folder_name):
                print('Clean up files in: {}/'.format(self.folder_name))
                shutil.rmtree(self.folder_name)
            print('Create the directory: {}/'.format(self.folder_name))
            os.mkdir(self.folder_name)

    def reset(self):
        self.fig = plt.figure(**self.fig_dict)

    def save(self, show=False, frame_count=None):
        if self.folder_name is None:
            self.fig.savefig(self.figure_name)
        else:
            frame_count = (
                self.frame_count if frame_count is None else frame_count
            )
            self.fig.savefig(
                self.folder_name + "/" +
                self.figure_name.format(frame_count)
            )
            self.frame_count += 1
        if show:
            plt.show()
        else:
            plt.close(self.fig)

    def movie(self, frame_rate, movie_name):
        print("Create movie:", movie_name+".mov")
        cmd = "ffmpeg -r {}".format(frame_rate)
        figure_name = self.figure_name.replace("{:", "%")
        figure_name = figure_name.replace("}", "")
        cmd += " -i " + self.folder_name + "/" + figure_name
        cmd += " -b:v 90M -c:v libx264 -pix_fmt yuv420p -f mov"
        cmd += " -y " + movie_name + ".mov"
        os.system(cmd)
